# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing Braket circuits used for testing

"""
import numpy as np
from braket.circuits import Circuit as BKCircuit


def braket_bell() -> BKCircuit:
    """Returns Braket bell circuit"""
    circuit = BKCircuit().h(0).cnot(0, 1)
    return circuit


def braket_shared15():
    """Returns braket `Circuit` for qBraid `TestSharedGates`."""

    circuit = BKCircuit()

    circuit.h([0, 1, 2, 3])
    circuit.x([0, 1])
    circuit.y(2)
    circuit.z(3)
    circuit.s(0)
    circuit.si(1)
    circuit.t(2)
    circuit.ti(3)
    circuit.rx(0, np.pi / 4)
    circuit.ry(1, np.pi / 2)
    circuit.rz(2, 3 * np.pi / 4)
    circuit.phaseshift(3, np.pi / 8)
    circuit.v(0)
    circuit.vi(1)
    circuit.iswap(2, 3)
    circuit.swap(0, 2)
    circuit.swap(1, 3)
    circuit.cnot(0, 1)
    circuit.cphaseshift(2, 3, np.pi / 4)

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module providing tools to map, analyze, and visualize conversion paths between different
quantum programs available through the qbraid.transpiler using directed graphs.

"""
from collections import deque
from importlib import import_module
from typing import Any, Callable, Optional, Union

import rustworkx as rx

from qbraid.programs.experiment import ExperimentType
from qbraid.programs.registry import (
    QPROGRAM_ALIASES,
    get_native_experiment_type,
    is_registered_alias_native,
)

from .edge import Conversion
from .exceptions import ConversionPathNotFoundError


def _get_path_from_bound_methods(bound_methods: list[Callable[..., Any]]) -> str:
    """
    Constructs a path string from a list of bound methods of Conversion instances.

    This function takes a list of bound methods (specifically 'convert' methods bound to
    Conversion instances) and constructs a path string representing the sequence of
    conversions. Each conversion is defined by the 'source' and 'target' properties of the
    Conversion instance to which each method is bound.

    Args:
        bound_methods: A list of bound methods from Conversion instances.

    Returns:
        A string representing the path of conversions, formatted as
        'source1 -> source2 -> ... -> targetN'.

    Raises:
        AttributeError: If the bound methods do not have the expected 'source'
                        and 'target' attributes.
        IndexError: If the list of bound methods is empty.
        TypeError: If an item in the bound_methods list is not a bound method.
    """
    if not bound_methods:
        raise IndexError("The list of bound methods is empty.")

    total_methods = len(bound_methods)

    path = []
    for index, method in enumerate(bound_methods):
        instance: Conversion = method.__self__  # Get the instance to which the method is bound
        if not hasattr(instance, "source") or not hasattr(instance, "target"):
            raise AttributeError("Bound method instance lacks 'source' or 'target' attributes.")
        path.append(instance.source)  # Add the source node of the instance

        # Add the target of the last method's instance to complete the path
        if index == total_methods - 1:
            path.append(instance.target)

    return " -> ".join(path)


def parse_conversion_path(conv_repr: str) -> list[tuple[str, str]]:
    """Parse a conversion path string into a list of conversion tuples."""
    components = conv_repr.split(" -> ")
    conversions = []
    for i in range(len(components) - 1):
        conversions.append((components[i], components[i + 1]))
    return conversions


# pylint: disable-next=too-many-public-methods
class ConversionGraph(rx.PyDiGraph):
    """
    Class for coordinating conversions between different quantum software programs

    """

    def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument
        return super().__new__(cls)  # pylint: disable=no-value-for-parameter

    # pylint: disable-next=too-many-arguments
    def __init__(
        self,
        conversions: Optional[list[Conversion]] = None,
        require_native: bool = False,
        include_isolated: bool = True,
        edge_bias: Optional[float] = None,
        nodes: Optional[Union[list[str], set[str]]] = None,
    ):
        """
        Initialize a ConversionGraph instance.

        Args:
            conversions (list[Conversion], optional): List of conversion edges. If None,
                default conversion edges are used.
            require_native (bool): If True, only include "native" conversion functions.
                Defaults to False.
            include_isolated (bool): If True, includes all registered program type aliases, even
                those that are not connected to any other nodes in the graph. Defaults to True.
            edge_bias (float, optional): Factor used to fine-tune the edge weight calculations
                and modify the decision thresholds for pathfinding. Defaults to 0.25 to prioritize
                shorter paths. For example, a bias of 0.25 slightly favors a single conversion at
                weight 0.78 over two conversions at weight 1.0. Only used if conversions is None.
            nodes (list[str], optional): List of nodes to include in the graph. If nodes is None
                and conversions is None, all nodes connected by registered conversions are included.
                If nodes is None and conversions is specified, only nodes connected by the specified
                conversions are included. Isolated nodes included iff include_isolated is True.

        """
        super().__init__()
        self.require_native = require_native
        self.edge_bias = edge_bias if edge_bias is not None else 0.25
        self._conversions = conversions or self.load_default_conversions(bias=self.edge_bias)
        self._node_alias_id_map: dict[str, int] = {}
        self._include_isolated = include_isolated
        self._init_nodes = set(nodes) if nodes is not None else set()
        self.create_conversion_graph()

    @staticmethod
    def load_default_conversions(bias: Optional[float] = None) -> list[Conversion]:
        """
        Create a list of default conversion nodes using predefined conversion functions.

        Returns:
            list[Conversion]: List of default conversion edges.
        """
        transpiler = import_module("qbraid.transpiler.conversions")
        conversion_functions: list[str] = getattr(transpiler, "conversion_functions", [])
        conversion_pairs: list[list[str, str]] = [
            conversion.split("_to_") for conversion in conversion_functions
        ]
        registered_conversion_pairs: list[list[str, str]] = [
            [source, target]
            for source, target in conversion_pairs
            if source in QPROGRAM_ALIASES and target in QPROGRAM_ALIASES
        ]

        def construct_conversion(name: list[str, str]) -> Conversion:
            source, target = name
            conversion_func = getattr(transpiler, f"{source}_to_{target}")
            return Conversion(source, target, conversion_func, bias=bias)

        return [construct_conversion(conversion) for conversion in registered_conversion_pairs]

    def create_conversion_graph(self) -> None:
        """Create a directed graph from a list of conversion functions."""
        nodes = self._init_nodes or set()

        for edge in (
            e for e in self._conversions if e.supported and (not self.require_native or e.native)
        ):
            if nodes and (edge.source not in nodes or edge.target not in nodes):
                continue
            if edge.source not in self._node_alias_id_map:
                self._node_alias_id_map[edge.source] = self.add_node(edge.source)
            if edge.target not in self._node_alias_id_map:
                self._node_alias_id_map[edge.target] = self.add_node(edge.target)
            self.add_edge(
                self._node_alias_id_map[edge.source],
                self._node_alias_id_map[edge.target],
                {"native": edge.native, "func": edge.convert, "weight": edge.weight},
            )

        if self._include_isolated:
            nodes = nodes or QPROGRAM_ALIASES
            for alias in nodes:
                if alias not in self._node_alias_id_map and (
                    not self.require_native or is_registered_alias_native(alias)
                ):
                    self._node_alias_id_map[alias] = self.add_node(alias)

    def has_node(self, node: str) -> bool:
        """
        Check if a node exists in the graph.

        Args:
            node (str): The node to check.

        Returns:
            bool: True if the node exists, False otherwise.
        """
        return node in set(self.nodes())

    def has_edge(self, node_a: str, node_b: str) -> bool:
        """
        Check if an edge exists between two nodes in the graph.

        Args:
            node_a (str): The source node.
            node_b (str): The target node.

        Returns:
            bool: True if the edge exists, False otherwise.
        """
        if not self.has_node(node_a) or not self.has_node(node_b):
            return False

        return super().has_edge(self._node_alias_id_map[node_a], self._node_alias_id_map[node_b])

    def conversions(self) -> list[Conversion]:
        """
        Get the list of conversion edges.

        Returns:
            list[Conversion]: The conversion edges of the graph.
        """
        return self._conversions

    def add_conversion(self, edge: Conversion, overwrite: bool = False) -> None:
        """
        Add a new conversion function as an edge in the graph.

        Args:
            edge (qbraid.interface.Conversion): The conversion edge to add to the graph.
            overwrite (optional, bool): If True, overwrites an existing conversion.
                                        Defaults to False.

        Raises:
            ValueError: If the conversion already exists and overwrite_existing is False.
        """
        source, target = edge.source, edge.target

        if self.has_edge(source, target) and not overwrite:
            raise ValueError(
                f"Conversion from {source} to {target} already exists. "
                "Set overwrite=True to overwrite."
            )

        for old_edge in self._conversions:
            if old_edge.source == source and old_edge.target == target:
                self._conversions.remove(old_edge)
                self.remove_edge(
                    self._node_alias_id_map[source],
                    self._node_alias_id_map[target],
                )
                break

        self._conversions.append(edge)

        if source not in self._node_alias_id_map:
            self._node_alias_id_map[source] = self.add_node(source)
        if target not in self._node_alias_id_map:
            self._node_alias_id_map[target] = self.add_node(target)
        self.add_edge(
            self._node_alias_id_map[source],
            self._node_alias_id_map[target],
            {"native": edge.native, "func": edge.convert, "weight": edge.weight},
        )

    def remove_conversion(self, source: str, target: str) -> None:
        """Safely remove a conversion from the graph."""
        if self.has_edge(source, target):
            self.remove_edge(self._node_alias_id_map[source], self._node_alias_id_map[target])
        else:
            raise ValueError(f"Conversion from {source} to {target} does not exist.")

        self._conversions = [
            conv
            for conv in self._conversions.copy()
            if not (conv.source == source and conv.target == target)
        ]

    def find_shortest_conversion_path(self, source: str, target: str) -> list[Callable]:
        """
        Find the shortest conversion path between two nodes in a graph.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.

        Returns:
            list of Callable: The shortest conversion path
                              as a list of bound methods of Conversion instances.

        Raises:
            ValueError: If no path is found between source and target.
        """
        path = rx.dijkstra_shortest_paths(
            self,
            self._node_alias_id_map[source],
            target=self._node_alias_id_map[target],
            weight_fn=lambda edge: edge["weight"],
        )

        if len(path) == 0:
            raise ConversionPathNotFoundError(source, target)

        return [
            self.get_edge_data(
                path[self._node_alias_id_map[target]][i],
                path[self._node_alias_id_map[target]][i + 1],
            )["func"]
            for i in range(len(path[self._node_alias_id_map[target]]) - 1)
        ]

    def find_top_shortest_conversion_paths(
        self, source: str, target: str, top_n: int = 3
    ) -> list[list[Callable]]:
        """
        Find the top shortest conversion paths between two nodes in a graph.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.
            top_n (int): Number of top shortest paths to find.

        Returns:
            list of list of Callable: The top shortest conversion paths.

        Raises:
            ConversionPathNotFoundError: If no path is found between source and target.
        """
        all_paths = rx.all_simple_paths(
            self, self._node_alias_id_map[source], self._node_alias_id_map[target]
        )

        # rx.all_simple_paths returns an empty list if no path is found
        if len(all_paths) == 0:
            raise ConversionPathNotFoundError(source, target)

        sorted_paths = sorted(all_paths, key=len)[:top_n]
        return [
            [self.get_edge_data(path[i], path[i + 1])["func"] for i in range(len(path) - 1)]
            for path in sorted_paths
        ]

    def has_path(self, source: str, target: str) -> bool:
        """
        Check if a conversion between two languages is supported.

        Args:
            source (str): The source language.
            target (str): The target language.

        Returns:
            bool: True if the conversion is supported, False otherwise.
        """
        if source == target:
            return True

        source_node = self._node_alias_id_map.get(source)
        target_node = self._node_alias_id_map.get(target)

        if source_node is None or target_node is None:
            return False

        return rx.has_path(self, source_node, target_node)

    def shortest_path(self, source: str, target: str) -> str:
        """
        Return string representation of the shortest conversion path between two nodes.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.

        Returns:
            str: String representation of shortest conversion path.

        Raises:
            ConversionPathNotFoundError: If no path is found between source and target.
        """
        path = self.find_shortest_conversion_path(source, target)
        return _get_path_from_bound_methods(path)

    def all_paths(self, source: str, target: str) -> list[str]:
        """
        Return string representations of all conversion paths between two nodes.

        Args:
            source (str): The starting node for the path.
            target (str): The target node for the path.

        Returns:
            list[str]: String representations of all conversion paths.

        Raises:
            ConversionPathNotFoundError: If no path is found between source and target.
        """
        num_conversions = len(self.conversions())
        paths = self.find_top_shortest_conversion_paths(source, target, top_n=num_conversions)
        return [_get_path_from_bound_methods(path) for path in paths]

    @staticmethod
    def _calculate_depth_and_weight(
        path: str, conv_weights: dict[tuple[str, str], int]
    ) -> tuple[int, int]:
        """Calculate the depth and weight of a given path."""
        conv_nodes = path.split(" -> ")
        conv_pairs = list(zip(conv_nodes, conv_nodes[1:]))
        depth = len(conv_pairs)
        weight = sum(conv_weights.get(pair, 0) for pair in conv_pairs)
        return depth, weight

    def _find_closest(
        self,
        pivot: str,
        candidates: list[str],
        has_path_func: Callable[[str, str], Optional[str]],
        shortest_path_func: Callable[[str, str], str],
    ) -> Optional[str]:
        """Find the closest item in candidates to/from pivot based on conversion paths."""
        closest = None
        min_depth = float("inf")
        min_weight = float("inf")

        conv_weights = {(conv.source, conv.target): conv.weight for conv in self.conversions()}

        for candidate in candidates:
            if has_path_func(pivot, candidate):
                path = shortest_path_func(pivot, candidate)
                depth, weight = self._calculate_depth_and_weight(path, conv_weights)

                if (depth < min_depth) or (depth == min_depth and weight < min_weight):
                    min_depth = depth
                    min_weight = weight
                    closest = candidate

        return closest

    def closest_target(self, source: str, targets: list[str]) -> Optional[str]:
        """
        Determine the closest target from a list of possible targets based on the
        shortest conversion path and weights. In case of equal path depths, the
        target with the higher total weight of its path edges is chosen.

        Args:
            source (str): The alias from which conversion paths are evaluated.
            targets (list[str]): A list of target aliases to which the conversion paths
                are evaluated.

        Returns:
            Optional[str]: The name of the target that requires the fewest conversions
                from the source or has higher weights in tie cases. Returns `None` if no
                conversion paths are available.
        """
        if source in targets:
            return source
        return self._find_closest(source, targets, self.has_path, self.shortest_path)

    def closest_source(self, target: str, sources: list[str]) -> Optional[str]:
        """
        Determine the closest source from a list of possible sources based on the
        shortest conversion path and weights. In case of equal path depths, the
        source with the higher total weight of its path edges is chosen.

        Args:
            target (str): The alias to which conversion paths are evaluated.
            sources (list[str]): A list of source aliases from which the conversion paths
                are evaluated.

        Returns:
            Optional[str]: The name of the source that requires the fewest conversions
                to the target or has higher weights in tie cases. Returns `None` if no
                conversion paths are available.
        """
        if target in sources:
            return target
        return self._find_closest(
            target,
            sources,
            lambda tgt, src: self.has_path(src, tgt),
            lambda tgt, src: self.shortest_path(src, tgt),
        )

    @staticmethod
    def _get_sorted_closest(
        pivot: str, items: list[str], closest_func: Callable[[str, list[str]], Optional[str]]
    ) -> list[str]:
        """Sort a list of items from closest to least close based on conversion paths."""
        sorted_items = []
        remaining = items.copy()

        while remaining:
            closest = closest_func(pivot, remaining)
            if closest is None:
                sorted_items.extend(remaining)
                break
            sorted_items.append(closest)
            remaining.remove(closest)

        return sorted_items

    def get_sorted_closest_targets(self, source: str, targets: list[str]) -> list[str]:
        """
        Sorts a list of targets from closest to least close based on conversion paths
        from the source. Targets without valid conversion paths are appended at the end.

        Args:
            source (str): The alias from which conversion paths are evaluated.
            targets (list[str]): A list of target aliases to be ordered by proximity.

        Returns:
            list[str]: A list of target aliases ordered by proximity to the source,
                    with unreachable targets appended at the end.
        """
        return self._get_sorted_closest(source, targets, self.closest_target)

    def get_sorted_closest_sources(self, target: str, sources: list[str]) -> list[str]:
        """
        Sorts a list of sources from closest to least close based on conversion paths
        from the target. Sources without valid conversion paths are appended at the end.

        Args:
            target (str): The alias to which conversion paths are evaluated.
            sources (list[str]): A list of source aliases to be ordered by proximity.

        Returns:
            list[str]: A list of source aliases ordered by proximity to the target,
                    with unreachable sources appended at the end.
        """
        return self._get_sorted_closest(target, sources, self.closest_source)

    def get_node_experiment_types(self) -> dict[str, ExperimentType]:
        """
        Get the experiment type of each node in the graph.

        Returns:
            dict[str, ExperimentType]: A dictionary mapping each node to its experiment type.
        """
        node_to_experiment_type = {}
        unassigned_nodes = deque()

        for node in self.nodes():
            if is_registered_alias_native(node):
                node_to_experiment_type[node] = get_native_experiment_type(node)
            else:
                node_to_experiment_type[node] = None
                unassigned_nodes.append(node)

        while unassigned_nodes:
            node = unassigned_nodes.popleft()
            assigned = False

            for recorded_node, recorded_exp_type in node_to_experiment_type.items():
                if recorded_exp_type is None or recorded_node is None:
                    continue

                if self.has_path(recorded_node, node) or self.has_path(node, recorded_node):
                    if node_to_experiment_type[node] is None:
                        node_to_experiment_type[node] = recorded_exp_type
                        assigned = True
                    elif node_to_experiment_type[node] != recorded_exp_type:
                        raise ValueError(
                            f"ExperimentType conflict detected: Node '{node}' "
                            f"(type '{node_to_experiment_type[node].name}') "
                            f"is reachable from node '{recorded_node}' "
                            f"(type '{recorded_exp_type.name}'), which has a different experiment "
                            "type. Ensure that connected nodes share compatible experiment types."
                        )

            if not assigned:
                node_to_experiment_type[node] = ExperimentType.OTHER

        return node_to_experiment_type

    def reset(self, conversions: Optional[list[Conversion]] = None) -> None:
        """
        Reset the graph to its default state.

        Returns:
            None
        """
        self.clear()
        self._conversions = conversions or self.load_default_conversions()
        self._node_alias_id_map = {}
        self.create_conversion_graph()

    def copy(self):
        """
        Create a copy of this graph, returning a new instance of ConversionGraph.

        """
        copied_conversions = self._conversions.copy() if self._conversions is not None else None
        return ConversionGraph(
            conversions=copied_conversions,
            require_native=self.require_native,
            include_isolated=self._include_isolated,
            edge_bias=self.edge_bias,
            nodes=self._init_nodes,
        )

    def subgraph(
        self, experiment_type: Union[ExperimentType, list[ExperimentType]]
    ) -> "ConversionGraph":
        """Filter the graph to include only nodes with the specified experiment type(s).

        Args:
            experiment_type (Union[ExperimentType, list[ExperimentType]]): The experiment type(s)

        Returns:
            ConversionGraph: A new ConversionGraph instance with only nodes of the
                specified experiment type(s).

        Raises:
            ValueError: If there are no nodes with the specified experiment type(s) in the graph.
        """
        node_experiment_types = self.get_node_experiment_types()
        exp_types = experiment_type if isinstance(experiment_type, list) else [experiment_type]
        nodes = [n for n in self.nodes() if node_experiment_types[n] in exp_types]

        if not nodes:
            exp_type_names = ", ".join([exp_type.name for exp_type in exp_types])
            raise ValueError(
                f"No program type nodes found with experiment type(s) '{exp_type_names}'. "
                "Use ConversionGraph.get_node_experiment_types() to inspect all experiment "
                "type mappings in this graph."
            )

        self._init_nodes = set(nodes)

        return self.copy()

    def __eq__(self, value: object) -> bool:
        return (
            super().__eq__(value)
            and isinstance(value, ConversionGraph)
            and self.conversions() == value.conversions()
            and self.require_native == value.require_native
            and self._node_alias_id_map == value._node_alias_id_map
            and self._include_isolated == value._include_isolated
            and self.edge_bias == value.edge_bias
            and self._init_nodes == value._init_nodes
        )

    def plot(self, **kwargs):
        """
        Plot the conversion graph.

        .. note::

            Matplotlib is an optional dependency and will not be installed with
            qbraid by default. If you intend to use this method make sure that
            you install matplotlib with either ``pip install matplotlib`` or
            ``pip install 'qbraid[visualization]'``.

        Args:
            **kwargs: Keyword arguments for the plot function.

        Returns:
            None
        """
        # pylint: disable-next=import-outside-toplevel
        from qbraid.visualization.plot_conversions import plot_conversion_graph

        plot_conversion_graph(self, **kwargs)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for converting Braket circuits to OpenQASM 3.0

"""
import pyqasm
from braket.circuits import Circuit
from braket.circuits.serialization import IRType

from qbraid.passes.qasm import replace_gate_names
from qbraid.transpiler.annotations import weight


@weight(1)
def braket_to_qasm3(circuit: Circuit) -> str:
    """Converts a ``braket.circuits.Circuit`` to an OpenQASM 3.0 string.

    Args:
        circuit: Amazon Braket quantum circuit

    Returns:
        The OpenQASM 3.0 string equivalent to the circuit

    Raises:
        ProgramConversionError: If braket to qasm conversion fails

    """
    has_measurements = any(str(instr.operator) == "Measure" for instr in circuit.instructions)
    qasm_program = circuit.to_ir(IRType.OPENQASM).source
    qasm_module = pyqasm.loads(qasm_program)
    qasm_module.unroll(external_gates=["cnot"])

    if not has_measurements:
        qasm_module.remove_measurements()

    qasm_str = pyqasm.dumps(qasm_module)
    qasm_str = replace_gate_names(qasm_str, {"cx": "cnot"})

    return qasm_str



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining function annotations (e.g. decorators) used in the transpiler.

"""

from functools import wraps
from typing import Any, Callable, TypeVar, cast

F = TypeVar("F", bound=Callable[..., Any])


def requires_extras(*packages: str) -> Callable[[F], F]:
    """
    Decorator factory to mark conversion functions that require additional dependencies
    beyond their "{source}_to_{target}" naming convention. It adds specified dependencies
    as attributes to the function.

    Args:
        *packages: The names of the required additional dependencies.

    Returns:
        Callable: A decorator that marks a function with the required dependencies.
    """

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        setattr(wrapper, "requires_extras", list(packages))
        return cast(F, wrapper)

    return decorator


def weight(value: float) -> Callable[[F], F]:
    """
    Decorator factory to mark conversion functions with a weight attribute.
    This weight attribute is used to prioritize conversion paths in a conversion graph.

    Args:
        value (float): The weight of the conversion function. Must be between 0 and 1 inclusive.

    Returns:
        Callable: A decorator that assigns the specified weight to a function.
    """
    if not 0 <= value <= 1:
        raise ValueError("Weight value must be between 0 and 1.")

    def decorator(func: F) -> F:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        setattr(wrapper, "weight", value)
        return cast(F, wrapper)

    return decorator



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Amazon Braket conversions

.. currentmodule:: qbraid.transpiler.conversions.braket

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   braket_to_cirq
   braket_to_qasm3
   braket_to_pytket
   braket_to_qiskit

"""
from .braket_extras import braket_to_pytket, braket_to_qiskit
from .braket_to_cirq import braket_to_cirq
from .braket_to_qasm3 import braket_to_qasm3

__all__ = [
    "braket_to_cirq",
    "braket_to_qasm3",
    "braket_to_pytket",
    "braket_to_qiskit",
]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Shared fixtures for Azure provider and testing circuit.
"""

import os
from typing import Optional

import pytest
from azure.identity import ClientSecretCredential
from azure.quantum import Workspace
from azure.quantum._constants import ConnectionConstants, EnvironmentVariables
from qiskit import QuantumCircuit

from qbraid.runtime.azure import AzureQuantumProvider


@pytest.fixture
def credential() -> Optional[ClientSecretCredential]:
    """Fixture for Azure client secret credential."""
    tenant_id = os.getenv(EnvironmentVariables.AZURE_TENANT_ID)
    client_id = os.getenv(EnvironmentVariables.AZURE_CLIENT_ID)
    client_secret = os.getenv(EnvironmentVariables.AZURE_CLIENT_SECRET)
    if client_id and tenant_id and client_secret:
        return ClientSecretCredential(
            tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
        )
    return None


@pytest.fixture
def resource_id() -> Optional[str]:
    """Fixture for Azure Quantum resource ID."""
    subscription_id = os.getenv(EnvironmentVariables.QUANTUM_SUBSCRIPTION_ID)
    resource_group = os.getenv(EnvironmentVariables.QUANTUM_RESOURCE_GROUP, "AzureQuantum")
    workspace_name = os.getenv(EnvironmentVariables.WORKSPACE_NAME)
    if subscription_id and resource_group and workspace_name:
        return ConnectionConstants.VALID_RESOURCE_ID(
            subscription_id=subscription_id,
            resource_group=resource_group,
            workspace_name=workspace_name,
        )
    return None


@pytest.fixture
def workspace(
    credential: Optional[ClientSecretCredential], resource_id: Optional[str]
) -> Workspace:
    """Fixture for Azure Quantum workspace."""
    location = os.getenv(EnvironmentVariables.QUANTUM_LOCATION, "eastus")
    return Workspace(resource_id=resource_id, location=location, credential=credential)


@pytest.fixture
def provider(workspace: Workspace) -> AzureQuantumProvider:
    """Fixture for AzureQuantumProvider."""
    return AzureQuantumProvider(workspace)


@pytest.fixture
def qiskit_circuit() -> QuantumCircuit:
    """Fixture for a Qiskit quantum circuit."""
    circuit = QuantumCircuit(3, 3)
    circuit.name = "main"
    circuit.h(0)
    circuit.cx(0, 1)
    circuit.cx(1, 2)
    circuit.measure([0, 1, 2], [0, 1, 2])

    return circuit



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for conversions between Cirq Circuits and QASM strings

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import cirq
import pyqasm
from cirq import ops, value

from qbraid._version import __version__ as qbraid_version
from qbraid.transpiler.annotations import weight

if TYPE_CHECKING:
    from qbraid.programs.typer import Qasm2StringType


@value.value_equality
class ZPowGate(cirq.ZPowGate):
    """A single qubit gate for rotations around the
    Z axis of the Bloch sphere.
    """

    def _qasm_(self, args: cirq.QasmArgs, qubits: tuple[cirq.Qid, ...]) -> Optional[str]:
        args.validate_version("2.0")
        if self._global_shift == 0:
            if self._exponent == 0.25:
                return args.format("t {0};\n", qubits[0])
            if self._exponent == -0.25:
                return args.format("tdg {0};\n", qubits[0])
            if self._exponent == 0.5:
                return args.format("s {0};\n", qubits[0])
            if self._exponent == -0.5:
                return args.format("sdg {0};\n", qubits[0])
            if self._exponent == 1:
                return args.format("z {0};\n", qubits[0])
            return args.format("p({0:half_turns}) {1};\n", self._exponent, qubits[0])
        return args.format("rz({0:half_turns}) {1};\n", self._exponent, qubits[0])


def map_zpow_and_unroll(circuit: cirq.Circuit) -> cirq.Circuit:
    """Map ZPowGate to RZ and unroll circuit"""

    def _map_zpow(op: cirq.Operation, _: int) -> cirq.OP_TREE:
        if isinstance(op.gate, cirq.ZPowGate):
            yield ZPowGate(exponent=op.gate.exponent, global_shift=op.gate.global_shift)(
                op.qubits[0]
            )
        else:
            yield op

    return cirq.map_operations_and_unroll(circuit, _map_zpow)


def _to_qasm_output(
    circuit: cirq.Circuit,
    header: Optional[str] = None,
    precision: int = 10,
    qubit_order: cirq.QubitOrderOrList = ops.QubitOrder.DEFAULT,
) -> cirq.QasmOutput:
    """Returns a QASM object equivalent to the circuit.

    Args:
        header: A multi-line string that is placed in a comment at the top
            of the QASM. Defaults to a cirq version specifier.
        precision: Number of digits to use when representing numbers.
        qubit_order: Determines how qubits are ordered in the QASM
            register.
    """
    if header is None:
        header = f"Generated from qBraid v{qbraid_version}"
    qubits = ops.QubitOrder.as_qubit_order(qubit_order).order_for(circuit.all_qubits())
    return cirq.QasmOutput(
        operations=circuit.all_operations(),
        qubits=qubits,
        header=header,
        precision=precision,
        version="2.0",
    )


@weight(1)
def cirq_to_qasm2(
    circuit: cirq.Circuit,
    header: Optional[str] = None,
    precision: int = 10,
    qubit_order: cirq.QubitOrderOrList = ops.QubitOrder.DEFAULT,
) -> Qasm2StringType:
    """Returns a QASM string representing the input Cirq circuit.

    Args:
        circuit: Cirq circuit to convert to a QASM string.

    Returns:
        Qasm2StringType: QASM string equivalent to the input Cirq circuit.
    """
    circuit = map_zpow_and_unroll(circuit)
    qasm = str(_to_qasm_output(circuit, header, precision, qubit_order))
    # format the qasm before returning
    return pyqasm.dumps(pyqasm.loads(qasm))



# Copyright (C) 2025 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining function to convert a CUDA-Q kernel to a QASM2 string.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import cudaq

from qbraid.transpiler.annotations import weight

if TYPE_CHECKING:
    from cudaq import PyKernel

    from qbraid.programs.typer import Qasm2StringType


@weight(1)
def cudaq_to_qasm2(kernel: PyKernel) -> Qasm2StringType:
    """Converts a CUDA-Q kernel to QASM2."""
    return cudaq.translate(kernel, format="openqasm2")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for managing and retrieving custom program type aliases

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from .exceptions import ProgramTypeError
from .exceptions import QasmError as QbraidQasmError
from .registry import QPROGRAM_REGISTRY, QPROGRAM_TYPES
from .typer import IonQDict, get_qasm_type_alias

if TYPE_CHECKING:
    import qbraid.programs


def find_str_type_alias(registry: dict[str, Type] = QPROGRAM_REGISTRY) -> Optional[str]:
    """Find additional keys with type 'str' in the registry."""
    str_keys = [
        k for k, v in registry.items() if v is str and k not in ("qasm2", "qasm3", "qasm2_kirin")
    ]

    if len(str_keys) == 0:
        return None
    if len(str_keys) == 1:
        return str_keys[0]
    raise ValueError(f"Multiple additional keys with type 'str' found: {str_keys}")


def _get_program_type_alias(program: qbraid.programs.QPROGRAM) -> str:
    """
    Get the type alias of a quantum program from registry.

    Args:
        program (qbraid.programs.QPROGRAM): The quantum program to get the type of.

    Returns:
        str: The type of the quantum program.

    Raises:
        ProgramTypeError: If the program type does not match any registered program types.
    """
    if isinstance(program, type):
        raise ProgramTypeError(message="Expected an instance of a quantum program, not a type.")

    if isinstance(program, str):
        try:
            return get_qasm_type_alias(program)
        except QbraidQasmError as err:
            package = find_str_type_alias()
            if package is not None:
                return package
            raise ProgramTypeError(
                message=(
                    "Input of program of type string does not represent a valid OpenQASM program, "
                    "and no additional string type aliases are registered."
                )
            ) from err

    if isinstance(program, IonQDict):
        return IonQDict.__alias__

    matched = []
    for alias, program_type in QPROGRAM_REGISTRY.items():
        if isinstance(program, (program_type, type(program_type))):
            matched.append(alias)

    if len(matched) == 1:
        return matched[0]

    if len(matched) > 1:
        raise ProgramTypeError(
            message=(
                f"Program of type '{type(program)}' matches multiple registered program types: "
                f"{matched}."
            )
        )

    raise ProgramTypeError(
        message=(
            f"Program of type '{type(program)}' does not match any registered "
            f"program types. Registered program types are: {QPROGRAM_TYPES}."
        )
    )


def get_program_type_alias(program: qbraid.programs.QPROGRAM, safe: bool = False) -> Optional[str]:
    """
    Get the type alias of a quantum program from registry.

    Args:
        program (qbraid.programs.QPROGRAM): The quantum program to get the type of.
        safe (bool): If True, return None if the program type does not match any registered
                     program types. Defaults to False.

    Returns:
        str: The type of the quantum program.
        None: If the program type does not match any registered program types and safe is True.

    Raises:
        ProgramTypeError: If the program type does not match any registered program types.
    """
    try:
        return _get_program_type_alias(program)
    except ProgramTypeError as err:
        if safe:
            return None
        raise err



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for caching module.

"""
import math

import pytest

from qbraid._caching import _generate_cache_key, cached_method, clear_cache


class TestClass:
    """Test class for cached_method decorator."""

    def __init__(self, x: int, y: int):
        self.x = x
        self.y = y

    @cached_method
    def adjusted_factorial(self, n: int) -> int:
        """Calculate the factorial of n, adjusted by the instance attrs x and y."""
        base_factorial = math.factorial(n + self.x)
        return base_factorial**self.y

    def __hash__(self):
        return hash((self.x, self.y))


@pytest.fixture
def test_instance():
    return TestClass(1, 2)


def test_generate_cache_key(test_instance):
    """Test cache key generation."""
    key = _generate_cache_key(test_instance, "get_data", (1,), {})
    assert isinstance(key, str)
    assert len(key) == 64  # SHA-256 hash length


def test_clear_cache(test_instance, monkeypatch):
    """Test clearing the LRU cache."""
    monkeypatch.setenv("DISABLE_CACHE", "0")

    assert test_instance.adjusted_factorial.cache_info().currsize == 0

    for i in range(3):
        test_instance.adjusted_factorial(i)

    assert test_instance.adjusted_factorial.cache_info().currsize == 3

    clear_cache()

    assert test_instance.adjusted_factorial.cache_info().currsize == 0



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for registering custom program types and aliases

"""
import sys
from typing import Any, Optional, Type, TypeVar, Union

from qbraid._entrypoints import get_entrypoints

from ._import import _QPROGRAM_ALIASES, _QPROGRAM_REGISTRY, _QPROGRAM_TYPES, NATIVE_REGISTRY
from .experiment import ExperimentType
from .typer import QbraidMetaType

QPROGRAM_NATIVE = Union[tuple(_QPROGRAM_TYPES)]
QPROGRAM_REGISTRY = _QPROGRAM_REGISTRY
QPROGRAM_ALIASES = _QPROGRAM_ALIASES
QPROGRAM_TYPES = _QPROGRAM_TYPES

QPROGRAM = TypeVar("QPROGRAM", bound=Any)


def derive_program_type_alias(program_type: Type[Any], use_submodule: bool = False) -> str:
    """
    Determines an alias for the given program type based on its module or class name.

    Args:
        program_type (Type[Any]): The class or type for which to determine the alias.
        use_submodule (bool): If True, alias will be based on first two parts of module name.

    Returns:
        str: The determined alias for the program type.

    Raises:
        ValueError: If the alias cannot be determined, or if use_submodule is True and the
                    module name is invalid or belongs to __main__ or builtins.
    """
    if isinstance(program_type, QbraidMetaType):
        return program_type.__alias__

    try:
        module_name = program_type.__module__
        module_parts = module_name.split(".")
        module_root = module_parts[0]

        if module_root in {"__main__", "builtins"}:
            if use_submodule:
                raise ValueError("Cannot use submodule aliasing with __main__ or builtins module.")
            alias = program_type.__name__
        else:
            if use_submodule:
                if len(module_parts) < 2:
                    raise ValueError(
                        "Module name does not have at least two parts for submodule aliasing."
                    )
                alias = f"{module_root}_{module_parts[1]}"
            else:
                alias = module_root

        return alias.lower()
    except (AttributeError, IndexError, TypeError) as err:
        raise ValueError(
            "Failed to automatically determine an alias from the program type's module."
        ) from err


def register_program_type(
    program_type: Type[Any], alias: Optional[str] = None, overwrite: bool = False
) -> None:
    """
    Registers a user-defined program type under the specified alias.

    Args:
        program_type (Type[Any]): The actual Python type or a callable that returns a type.
                             This can be a built-in type like str, a class, or any callable.
        alias (optional, str): The alias to register the program type under.
        overwrite (optional, bool): Whether to overwrite an existing alias with the new type.

    Raises:
        ValueError: If the alias is already registered with a different type,
                    if the program type is already registered under a different alias,
                    or if trying to register more than one additional 'str' type beyond
                    'qasm2', 'qasm3', and 'qasm2_kirin'.
    """
    if not alias:
        alias = derive_program_type_alias(program_type)

    normalized_alias = alias.lower()

    # Check if the alias is already used and if it maps to a different type
    if normalized_alias in QPROGRAM_REGISTRY:
        registered_type = QPROGRAM_REGISTRY[normalized_alias]
        if registered_type != program_type and overwrite is False:
            if (
                isinstance(registered_type, QbraidMetaType)
                and program_type == registered_type.__bound__
            ):
                program_type = registered_type
            else:
                raise ValueError(f"Alias '{alias}' is already registered with a different type.")

    # Check if the type is already registered under any other alias
    existing_alias = next((k for k, v in QPROGRAM_REGISTRY.items() if v == program_type), None)
    if existing_alias and existing_alias != normalized_alias and overwrite is False:
        if program_type is str:
            str_types = [
                k
                for k, v in QPROGRAM_REGISTRY.items()
                if v is str and k not in ("qasm2", "qasm3", "qasm2_kirin")
            ]
            if (
                len(str_types) >= 1
                and normalized_alias not in str_types
                and normalized_alias not in ("qasm2", "qasm3", "qasm2_kirin")
            ):
                raise ValueError(
                    "Cannot register more than one additional 'str' type beyond "
                    "'qasm2', 'qasm3', and 'qasm2_kirin'."
                )
        else:
            raise ValueError(
                f"Program type '{program_type}' already registered under alias '{existing_alias}'."
            )

    # Register the new type and alias
    QPROGRAM_REGISTRY[normalized_alias] = program_type
    QPROGRAM_ALIASES.add(normalized_alias)
    QPROGRAM_TYPES.add(program_type)


def unregister_program_type(alias: str, raise_error: bool = True) -> None:
    """
    Unregisters the program type associated with the given alias.

    Args:
        alias (str): The alias to unregister.
        raise_error (bool): If True, raises KeyError if the alias is not found. Defaults to True.

    Raises:
        KeyError: If the alias is not registered and raise_error is True.
    """
    normalized_alias = alias.lower()

    QPROGRAM_ALIASES.discard(normalized_alias)

    if normalized_alias not in QPROGRAM_REGISTRY:
        if raise_error:
            raise KeyError(f"No program type registered under the alias '{alias}'.")
        return

    program_type = QPROGRAM_REGISTRY.pop(normalized_alias)

    if not any(pt == program_type for pt in QPROGRAM_REGISTRY.values()):
        QPROGRAM_TYPES.discard(program_type)


def is_registered_alias_native(alias: str) -> bool:
    """
    Determine if the registered program type for a given alias matches the native program type.

    Args:
        alias (str): The alias to check against the native and registered program types.

    Returns:
        bool: True if the alias is registered and its program type matches the native type,
              otherwise False.
    """
    native_type = NATIVE_REGISTRY.get(alias)
    registered_type = QPROGRAM_REGISTRY.get(alias)

    return native_type is not None and native_type == registered_type


def get_native_experiment_type(native_alias: str) -> ExperimentType:
    """Returns the ExperimentType corresponding to the native program type alias."""
    module = "programs"
    group = f"qbraid.{module}"

    native_no_ep = {
        "openqasm3": ExperimentType.GATE_MODEL,
        "qasm2_kirin": ExperimentType.GATE_MODEL,
        "pyqir": ExperimentType.GATE_MODEL,
        "stim": ExperimentType.GATE_MODEL,
        "qibo": ExperimentType.GATE_MODEL,
        "bloqade": ExperimentType.AHS,
        "pulser": ExperimentType.AHS,
    }

    if native_alias in native_no_ep:
        return native_no_ep[native_alias]

    entry_points = get_entrypoints(module)
    entry_point = entry_points.get(native_alias)

    if entry_point is None:
        raise ValueError(f"Entry point '{native_alias}' not found in '{group}'.")

    module_path = (
        entry_point.value.split(":")[0] if sys.version_info >= (3, 10) else entry_point.module_name
    )

    programs_sub_module = module_path[len(group) + 1 :].split(".", maxsplit=1)[0]

    return ExperimentType(programs_sub_module)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for defining and updating runtime conversion schemes

"""

import pytest
import rustworkx as rx

from qbraid.programs.spec import ProgramSpec
from qbraid.transpiler.graph import ConversionGraph
from qbraid.transpiler.scheme import ConversionScheme


def test_initialization():
    """Test the initialization and default values of the ConversionScheme."""
    cs = ConversionScheme()
    assert cs.conversion_graph is None
    assert cs.max_path_attempts == 3
    assert cs.max_path_depth is None
    assert len(cs.extra_kwargs) == 0


def test_update_values():
    """Test the dynamic update of the ConversionScheme's attributes."""
    cs = ConversionScheme()
    cs.update_values(max_path_attempts=5, max_path_depth=2, extra_kwargs={"new_key": "new_value"})
    assert cs.max_path_attempts == 5
    assert cs.max_path_depth == 2
    assert cs.extra_kwargs == {"new_key": "new_value"}

    # Test updating nested kwargs
    cs.update_values(extra_kwargs={"new_key": "updated_value"})
    assert cs.extra_kwargs["new_key"] == "updated_value"


def test_update_invalid_attribute():
    """Test updating an invalid attribute which should raise an AttributeError."""
    cs = ConversionScheme()
    with pytest.raises(AttributeError):
        cs.update_values(invalid_key="value")


def test_str_method():
    """Test the string representation of the ConversionScheme."""
    cs = ConversionScheme(
        conversion_graph=None,
        max_path_attempts=2,
        max_path_depth=1,
        extra_kwargs={"require_native": True},
    )
    expected_str = (
        "ConversionScheme(conversion_graph=None, max_path_attempts=2, "
        "max_path_depth=1, require_native=True)"
    )
    assert str(cs) == expected_str


@pytest.fixture
def rx_graph():
    """Returns a rustworkx.PyDiGraph for testing."""
    graph = rx.PyDiGraph()
    a = graph.add_node("A")
    b = graph.add_node("B")
    c = graph.add_node("C")
    d = graph.add_node("D")
    e = graph.add_node("E")
    graph.add_edges_from(
        [(a, b, 0.1), (b, c, 1.2), (c, a, 2.0), (d, c, 1.0), (e, d, 2.1), (e, b, 0.2)]
    )
    return graph


@pytest.mark.parametrize(
    "target_node, max_edges, expected_result",
    [
        (["C"], 1, {"C", "B", "D"}),
        (["C"], 2, {"C", "B", "D", "E", "A"}),
        (["A"], 2, {"A", "B", "C", "D"}),
        (["E"], 0, {"E"}),
        (["E", "C"], 1, {"E", "C", "B", "D"}),
        (["A"], None, {"A", "B", "C", "D", "E"}),
    ],
)
def test_find_nodes_reachable_within_max_edges(rx_graph, target_node, max_edges, expected_result):
    """Test finding all nodes reachable from a target node within a specified number of edges."""
    result = ConversionScheme.find_nodes_reachable_within_max_edges(
        rx_graph, target_node, max_edges
    )
    assert result == expected_result


def test_find_nodes_reachable_within_max_edges_raises_for_invalid_target(rx_graph):
    """Test that an error is raised when the target node is not in the graph."""
    with pytest.raises(ValueError) as excinfo:
        ConversionScheme.find_nodes_reachable_within_max_edges(rx_graph, ["F"], 1)
    assert "Target node 'F' not found in the graph." in str(excinfo.value)


def test_update_graph_for_target():
    """Test updating the conversion graph to include only nodes with paths to the target node(s)."""
    qasm2_spec = ProgramSpec(str, alias="qasm2")
    qasm3_spec = ProgramSpec(str, alias="qasm3")

    target_spec = [qasm2_spec, qasm3_spec]

    graph = ConversionGraph(include_isolated=True)

    qasm_nodes = set()
    non_qasm_nodes = set()
    for node in graph.nodes():
        if graph.has_path(node, qasm2_spec.alias) or graph.has_path(node, qasm3_spec.alias):
            qasm_nodes.add(node)
        else:
            non_qasm_nodes.add(node)

    scheme = ConversionScheme(conversion_graph=graph)

    scheme.update_graph_for_target(target_spec)

    updated_graph = scheme.conversion_graph
    assert isinstance(updated_graph, ConversionGraph)

    original_nodes = set(graph.nodes())
    updated_nodes = set(updated_graph.nodes())
    assert updated_nodes == qasm_nodes
    assert len(updated_nodes) == len(original_nodes) - len(non_qasm_nodes)


def test_find_nodes_reachable_within_max_edges_raises_for_negative(rx_graph):
    """Test that an error is raised when the max_edges is negative."""
    with pytest.raises(
        ValueError, match="The maximum number of edges must be a non-negative integer."
    ):
        ConversionScheme.find_nodes_reachable_within_max_edges(rx_graph, ["A"], -1)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for qbraid.programs.braket.BraketCircuit

"""
from itertools import chain, combinations
from unittest.mock import Mock

import numpy as np
import pytest
from braket.circuits import Circuit, Instruction, gates

from qbraid.programs import ProgramTypeError
from qbraid.programs.gate_model.braket import BraketCircuit


def get_subsets(nqubits):
    """Return list of all combinations up to number nqubits"""
    qubits = list(range(0, nqubits))

    def combos(x):
        return combinations(qubits, x)

    all_subsets = chain(*map(combos, range(0, len(qubits) + 1)))
    return list(all_subsets)[1:]


def calculate_expected(gates_set):
    """Calculated expected unitary"""
    if len(gates_set) == 1:
        return gates_set[0]
    if len(gates_set) == 2:
        return np.kron(gates_set[1], gates_set[0])
    return np.kron(calculate_expected(gates_set[2:]), np.kron(gates_set[1], gates_set[0]))


def generate_test_data(input_gate_set, contiguous=True):
    """Generate test data"""
    testdata = []
    gate_set = input_gate_set.copy()
    gate_set.append(gates.I)
    nqubits = len(input_gate_set)
    subsets = get_subsets(nqubits)
    for ss in subsets:
        bk_instrs = []
        np_gates = []
        for index in range(max(ss) + 1):
            idx = -1 if index not in ss else index
            BKGate = gate_set[idx]
            np_gate = BKGate().to_matrix()
            if idx != -1 or contiguous:
                bk_instrs.append((BKGate, index))
            np_gates.append(np_gate)
        u_expected = calculate_expected(np_gates)
        testdata.append((bk_instrs, u_expected))
    return testdata


def make_circuit(bk_instrs):
    """Constructs Braket circuit from list of instructions"""
    circuit = Circuit()
    for instr in bk_instrs:
        Gate, index = instr
        circuit.add_instruction(Instruction(Gate(), target=index))
    qprogram = BraketCircuit(circuit)
    qprogram.populate_idle_qubits()
    return qprogram.program


test_gate_set = [gates.X, gates.Y, gates.Z]
test_data_contiguous_qubits = generate_test_data(test_gate_set)
test_data_non_contiguous_qubits = generate_test_data(test_gate_set, contiguous=False)
test_data = test_data_contiguous_qubits + test_data_non_contiguous_qubits


@pytest.mark.parametrize("bk_instrs,u_expected", test_data)
def test_unitary_calc(bk_instrs, u_expected):
    """Test calculating unitary of circuits using both contiguous and
    non-contiguous qubit indexing."""
    circuit = make_circuit(bk_instrs)
    qbraid_circuit = BraketCircuit(circuit)
    u_test = qbraid_circuit.unitary_little_endian()
    assert np.allclose(u_expected, u_test)


@pytest.mark.parametrize("bk_instrs,u_expected", test_data)
def test_convert_be_to_le(bk_instrs, u_expected):
    """Test converting big-endian unitary to little-endian unitary."""
    circuit = make_circuit(bk_instrs)
    qbraid_circuit = BraketCircuit(circuit)
    u_test = qbraid_circuit.unitary_little_endian()
    assert np.allclose(u_expected, u_test)


def test_kronecker_product_factor_permutation():
    """Test calculating unitary permutation representing
    circuits with reversed qubits"""
    bk_circuit = Circuit().h(0).cnot(0, 1)
    circuit = BraketCircuit(bk_circuit)

    unitary = circuit.unitary()
    circuit.reverse_qubit_order()
    unitary_rev = circuit.unitary_rev_qubits()

    assert np.allclose(unitary, unitary_rev)


def test_unitary_little_endian_braket_bell():
    """Test convert_to_contigious on bell circuit"""
    circuit = Circuit().h(0).cnot(0, 1)  # pylint: disable=no-member
    h_gate = np.sqrt(1 / 2) * np.array([[1, 1], [1, -1]])
    h_gate_kron = np.kron(np.eye(2), h_gate)
    cnot_gate = np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])
    u_expected = np.einsum("ij,jk->ki", h_gate_kron, cnot_gate)
    qprogram = BraketCircuit(circuit)
    qprogram.remove_idle_qubits()
    u_little_endian = qprogram.unitary_little_endian()
    assert np.allclose(u_expected, u_little_endian)


def test_collapse_empty_braket_control_modifier():
    """Test that converting braket circuits to contiguous qubits works with control modifiers"""
    circuit = Circuit().y(target=0, control=1)
    qprogram = BraketCircuit(circuit)
    qprogram.remove_idle_qubits()
    contig_circuit = qprogram.program
    assert circuit.qubit_count == contig_circuit.qubit_count


def test_raise_program_type_error():
    """Test raising ProgramTypeError"""
    with pytest.raises(ProgramTypeError):
        BraketCircuit(Mock())


def test_properties():
    """Test properties of BraketCircuit"""
    circuit = Circuit().h(0).cnot(0, 1)
    qprogram = BraketCircuit(circuit)
    assert qprogram.qubits == circuit.qubits
    assert qprogram.num_clbits == 0
    assert qprogram.depth == circuit.depth


def test_pad_measurements_no_existing_measurements():
    """Test pad_measurements method when no measurements exist."""
    circuit = Circuit().h(0).cnot(0, 1)
    qprogram = BraketCircuit(circuit)
    qprogram.pad_measurements()

    # Should have no measurement
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 0


def test_pad_measurements_with_existing_measurements():
    """Test pad_measurements method when some measurements already exist."""
    circuit = Circuit().h(0).cnot(0, 1).measure(0)
    qprogram = BraketCircuit(circuit)
    qprogram.pad_measurements()

    # Should have measurements on all qubits now
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 2

    # Should track the originally measured qubit
    assert hasattr(qprogram.program, "partial_measurement_qubits")
    assert qprogram.program.partial_measurement_qubits == [0]


def test_pad_measurements_multiple_partial_measurements():
    """Test pad_measurements method with multiple partial measurements."""
    circuit = Circuit().h(0).cnot(0, 1).cnot(1, 2).measure(0).measure(2)
    qprogram = BraketCircuit(circuit)
    qprogram.pad_measurements()

    # Should have measurements on all qubits now
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 3

    # Should track the originally measured qubits
    assert hasattr(qprogram.program, "partial_measurement_qubits")
    assert set(qprogram.program.partial_measurement_qubits) == {0, 2}


def test_transform_with_ionq_device():
    """Test transform method with IonQ device calls pad_measurements."""
    circuit = Circuit().h(0).cnot(0, 1).measure(0)
    qprogram = BraketCircuit(circuit)

    # Mock device with IonQ provider
    mock_device = Mock()
    mock_device.simulator = False
    mock_device._provider_name = "IonQ"

    qprogram.transform(mock_device)

    # Should have padded measurements
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 2
    assert hasattr(qprogram.program, "partial_measurement_qubits")
    assert qprogram.program.partial_measurement_qubits == [0]


def test_transform_with_amazon_braket_device():
    """Test transform method with Amazon Braket device calls pad_measurements."""
    circuit = Circuit().h(0).cnot(0, 1).measure(1)
    qprogram = BraketCircuit(circuit)

    # Mock device with Amazon Braket provider
    mock_device = Mock()
    mock_device.simulator = False
    mock_device._provider_name = "Amazon Braket"

    qprogram.transform(mock_device)

    # Should have padded measurements
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 2
    assert hasattr(qprogram.program, "partial_measurement_qubits")
    assert qprogram.program.partial_measurement_qubits == [1]


def test_transform_with_other_provider():
    """Test transform method with other provider doesn't call pad_measurements."""
    circuit = Circuit().h(0).cnot(0, 1).measure(0)
    qprogram = BraketCircuit(circuit)

    # Mock device with other provider
    mock_device = Mock()
    mock_device.simulator = False
    mock_device._provider_name = "Other Provider"

    qprogram.transform(mock_device)

    # Should not have padded measurements
    measurement_count = sum(
        1
        for instr in qprogram.program.instructions
        if hasattr(instr.operator, "__class__") and instr.operator.__class__.__name__ == "Measure"
    )
    assert measurement_count == 1
    assert not hasattr(qprogram.program, "partial_measurement_qubits")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit tests for IonQ runtime (remote)

"""

import os
from typing import TYPE_CHECKING

import pytest
from qbraid_core.exceptions import RequestsApiError

from qbraid._logging import logger
from qbraid.runtime import GateModelResultData, IonQDevice, IonQProvider, JobStateError, Result

if TYPE_CHECKING:
    import cirq as cirq_module
    import qiskit as qiskit_module


@pytest.mark.remote
def test_ionq_multicircuit_job():
    """Test running multiple circuits in a single job."""
    cirq: cirq_module = pytest.importorskip("cirq")
    qiskit: qiskit_module = pytest.importorskip("qiskit")

    api_key = os.getenv("IONQ_API_KEY")

    if not api_key:
        pytest.skip("IONQ_API_KEY is not set")

    provider = IonQProvider(api_key=api_key)
    device = provider.get_device("simulator")

    qiskit_ghz = qiskit.QuantumCircuit(3)
    qiskit_ghz.h(0)
    qiskit_ghz.cx(0, 1)
    qiskit_ghz.cx(0, 2)

    cirq_bell = cirq.Circuit()
    q0, q1 = cirq.LineQubit.range(2)
    cirq_bell.append(cirq.H(q0))
    cirq_bell.append(cirq.CNOT(q0, q1))

    device = provider.get_device("simulator")

    assert isinstance(device, IonQDevice)

    job = device.run([qiskit_ghz, cirq_bell], name="qBraid Integration Test", shots=1000)

    try:
        job.wait_for_final_state(timeout=60)
    except TimeoutError as err:
        logger.error(err)

        try:
            job.cancel()
        except (RequestsApiError, JobStateError) as err:
            logger.error(err)

        pytest.skip(reason="Job did not complete within the timeout")

    result = job.result()

    assert isinstance(result, Result)

    result_data = result.data

    assert isinstance(result_data, GateModelResultData)

    counts = result_data.get_counts()

    assert counts == [{"000": 500, "111": 500}, {"000": 500, "011": 500}]



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=unused-import

"""
Fixtures imported/defined in this file can be used by any test in this directory
without needing to import them (pytest will automatically discover them).

"""
import importlib.util
import os
import sys

import pytest

from .fixtures import bell_circuit as bell_circuit
from .fixtures import bell_unitary as bell_unitary
from .fixtures import packages_bell as packages_bell
from .fixtures import packages_shared15 as packages_shared15
from .fixtures import shared15_circuit as shared15_circuit
from .fixtures import shared15_unitary as shared15_unitary
from .fixtures import two_bell_circuits as two_bell_circuits
from .fixtures import two_shared15_circuits as two_shared15_circuits


def _is_package_installed(package_name: str) -> bool:
    """Check if a package is installed."""
    return importlib.util.find_spec(package_name) is not None


@pytest.fixture
def available_targets():
    """Return a list of available quantum packages."""
    packages = ["braket", "cirq", "pytket", "qiskit", "pyquil"]

    # Add version-restricted packages only if Python version is compatible
    if sys.version_info < (3, 12):
        packages.extend(["pyqubo"])
    if sys.version_info < (3, 13):
        packages.extend(["bloqade"])

    return [pkg for pkg in packages if _is_package_installed(pkg)]


def pytest_addoption(parser):
    """Adds custom remote testing command-line option to pytest."""
    parser.addoption(
        "--remote",
        action="store",
        default=None,
        help="Run tests that interface with remote, credentialed services: true or false",
    )


def pytest_collection_modifyitems(config, items):
    """Skip tests marked with `remote` if remote tests are disabled."""
    remote_option = config.getoption("--remote")
    if remote_option is None:
        remote_option = os.getenv("QBRAID_RUN_REMOTE_TESTS", "False").lower() == "true"
    else:
        remote_option = remote_option.lower() == "true"

    if not remote_option:
        skip_remote = pytest.mark.skip(reason="Remote tests are disabled.")
        for item in items:
            if "remote" in item.keywords:
                item.add_marker(skip_remote)


@pytest.fixture(autouse=True)
def disable_cache_for_tests(monkeypatch):
    """Disable caching for all tests."""
    monkeypatch.setenv("DISABLE_CACHE", "1")



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

# pylint: disable=redefined-outer-name

"""
Unit tests for BraketProvider class

"""
import datetime
import importlib.util
import json
import warnings
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
from botocore.exceptions import NoCredentialsError
from braket.aws.aws_session import AwsSession
from braket.aws.queue_information import QueueDepthInfo, QueueType
from braket.circuits import Circuit
from braket.device_schema import ExecutionDay
from braket.devices import Devices, LocalSimulator

from qbraid.exceptions import QbraidError
from qbraid.interface import circuits_allclose
from qbraid.programs import ExperimentType, ProgramSpec, load_program
from qbraid.runtime import (
    DeviceProgramTypeMismatchError,
    DeviceStatus,
    GateModelResultData,
    TargetProfile,
)
from qbraid.runtime.aws.availability import _calculate_future_time
from qbraid.runtime.aws.device import BraketDevice
from qbraid.runtime.aws.job import AmazonBraketVersionError, BraketQuantumTask
from qbraid.runtime.aws.provider import BraketProvider
from qbraid.runtime.aws.result_builder import BraketGateModelResultBuilder
from qbraid.runtime.exceptions import JobStateError, ProgramValidationError

from ...fixtures.braket.gates import get_braket_gates

braket_gates = get_braket_gates()

pytket_not_installed = importlib.util.find_spec("pytket") is None

SV1_ARN = Devices.Amazon.SV1


@pytest.fixture
def braket_provider():
    """Return a BraketProvider instance."""
    return BraketProvider()


@pytest.fixture
def sv1_profile():
    """Target profile for AWS SV1 device."""
    return TargetProfile(
        simulator=True,
        num_qubits=34,
        program_spec=ProgramSpec(Circuit),
        provider_name="Amazon Braket",
        device_id=SV1_ARN,
        experiment_type=ExperimentType.GATE_MODEL,
    )


class TestAwsSession:
    """Test class for AWS session."""

    def __init__(self):
        self.region = "us-east-1"

    def get_device(self, arn):  # pylint: disable=unused-argument
        """Returns metadata for a device."""
        capabilities = {
            "action": {
                "braket.ir.openqasm.program": "literally anything",
            },
            "paradigm": {
                "qubitCount": 2,
                "nativeGateSet": ["gate1", "gate2"],
            },
        }
        cap_json = json.dumps(capabilities)
        metadata = {
            "deviceType": "SIMULATOR",
            "providerName": "Amazon Braket",
            "deviceCapabilities": cap_json,
        }

        return metadata

    def cancel_quantum_task(self, arn):  # pylint: disable=unused-argument
        """Cancel a quantum task."""
        return None


class ExecutionWindow:
    """Test class for execution window."""

    def __init__(self):
        self.windowStartHour = datetime.time(0)
        self.windowEndHour = datetime.time(23, 59, 59)
        self.executionDay = ExecutionDay.EVERYDAY


class Service:
    """Test class for braket device service."""

    def __init__(self):
        self.executionWindows = [ExecutionWindow()]


class TestProperties:
    """Test class for braket device properties."""

    def __init__(self):
        self.service = Service()


class TestAwsDevice:
    """Test class for braket device."""

    def __init__(self, arn, aws_session=None):
        self.arn = arn
        self.name = "SV1"
        self.aws_session = aws_session or TestAwsSession()
        self.status = "ONLINE"
        self.properties = TestProperties()
        self.is_available = True

    @staticmethod
    def get_device_region(arn):  # pylint: disable=unused-argument
        """Returns the region of a device."""
        return "us-east-1"


class MockTask:
    """Mock task class."""

    def __init__(self, arg):
        self.id = arg
        self.arn = arg
        self._aws_session = TestAwsSession()

    def result(self):
        """Mock result method."""
        return "not a result"

    def state(self):
        """Mock state method."""
        return "COMPLETED" if self.id == "task1" else "RUNNING"

    def cancel(self):
        """Mock cancel method."""
        raise RuntimeError

    def metadata(self):
        """Mock metadata method."""
        return {"status": self.state(), "quantumTaskArn": self.id, "deviceArn": SV1_ARN}


@pytest.fixture
def mock_sv1():
    """Return a mock SV1 device."""
    return TestAwsDevice(SV1_ARN)


@pytest.fixture
def mock_aws_configure():
    """Mock aws_conifugre function."""
    with patch("qbraid.runtime.aws.provider.aws_configure") as mock:
        yield mock


@pytest.mark.parametrize(
    "input_keys, expected_calls",
    [
        (
            {"aws_access_key_id": "AKIA...", "aws_secret_access_key": "secret"},
            {"aws_access_key_id": "AKIA...", "aws_secret_access_key": "secret"},
        ),
        ({}, {"aws_access_key_id": "default_key", "aws_secret_access_key": "default_secret"}),
        (
            {"aws_access_key_id": "AKIA..."},
            {"aws_access_key_id": "AKIA...", "aws_secret_access_key": "default_secret"},
        ),
    ],
)
def test_save_config(mock_aws_configure, input_keys, expected_calls):
    """Test save_config method of BraketProvider class."""
    instance = BraketProvider()
    instance.aws_access_key_id = "default_key"
    instance.aws_secret_access_key = "default_secret"

    instance.save_config(**input_keys)

    mock_aws_configure.assert_called_once_with(**expected_calls)


def test_provider_get_aws_session():
    """Test getting an AWS session."""
    with patch("boto3.session.Session") as mock_boto_session:
        mock_boto_session.aws_access_key_id.return_value = "aws_access_key_id"
        mock_boto_session.aws_secret_access_key.return_value = "aws_secret_access_key"
        aws_session = BraketProvider()._get_aws_session()
        assert aws_session.boto_session.region_name == "us-east-1"
        assert isinstance(aws_session, AwsSession)


def test_provider_get_aws_session_with_session_token():
    """Test getting an AWS session with session token."""
    with patch("boto3.session.Session") as mock_boto_session:
        mock_boto_session.aws_access_key_id.return_value = "aws_access_key_id"
        mock_boto_session.aws_secret_access_key.return_value = "aws_secret_access_key"
        mock_boto_session.get_credentials.return_value.token = "session_token"
        aws_session = BraketProvider(aws_session_token="session_token")._get_aws_session()
        assert aws_session.boto_session.region_name == "us-east-1"
        assert isinstance(aws_session, AwsSession)
        assert aws_session.boto_session.get_credentials().token == "session_token"


def test_provider_get_devices_raises(braket_provider):
    """Test raising device wrapper error due to invalid device ID."""
    with pytest.raises(ValueError):
        braket_provider.get_device("Id123")


def test_provider_build_runtime_profile(mock_sv1):
    """Test building a runtime profile."""
    provider = BraketProvider()
    profile = provider._build_runtime_profile(device=mock_sv1, extra="data")
    assert profile.get("simulator") is True
    assert profile.get("provider_name") == "Amazon Braket"
    assert profile.get("device_id") == SV1_ARN
    assert profile.get("extra") == "data"
    assert profile.get("basis_gates") == {"gate1", "gate2"}


@pytest.mark.parametrize(
    "region_names, key, values, expected",
    [(["us-west-1"], "Project", ["X"], ["arn:aws:resource1", "arn:aws:resource2"])],
)
@patch("boto3.client")
def test_provider_get_tasks_by_tag(mock_boto_client, region_names, key, values, expected):
    """Test fetching tagged resources from AWS."""
    mock_client = MagicMock()
    mock_boto_client.return_value = mock_client
    mock_client.get_resources.return_value = {
        "ResourceTagMappingList": [
            {"ResourceARN": "arn:aws:resource1"},
            {"ResourceARN": "arn:aws:resource2"},
        ]
    }

    provider = BraketProvider()
    result = provider.get_tasks_by_tag(key, values, region_names)

    # Assert the results
    mock_boto_client.assert_called_with("resourcegroupstaggingapi", region_name="us-west-1")
    mock_client.get_resources.assert_called_once_with(TagFilters=[{"Key": key, "Values": values}])
    assert result == expected


def test_provider_get_device(mock_sv1):
    """Test getting a Braket device."""
    provider = BraketProvider()
    with (
        patch("qbraid.runtime.aws.provider.AwsDevice") as mock_aws_device,
        patch("qbraid.runtime.aws.device.AwsDevice") as mock_aws_device_2,
    ):
        mock_aws_device.return_value = mock_sv1
        mock_aws_device_2.return_value = mock_sv1
        device = provider.get_device(SV1_ARN)
        assert device.id == SV1_ARN
        assert device.name == "SV1"
        assert str(device) == "BraketDevice('Amazon Braket SV1')"
        assert device.status() == DeviceStatus.ONLINE
        assert isinstance(device, BraketDevice)


def test_provider_get_devices(mock_sv1):
    """Test getting list of Braket devices."""
    with (
        patch("qbraid.runtime.aws.provider.AwsDevice.get_devices", return_value=[mock_sv1]),
        patch("qbraid.runtime.aws.device.AwsDevice") as mock_aws_device_2,
    ):
        provider = BraketProvider()
        mock_aws_device_2.return_value = mock_sv1
        devices = provider.get_devices()
        assert len(devices) == 1
        assert devices[0].id == SV1_ARN


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_profile_attributes(mock_aws_device, sv1_profile):
    """Test that device profile attributes are correctly set."""
    mock_aws_device.return_value = Mock()
    device = BraketDevice(sv1_profile)
    assert device.id == sv1_profile.get("device_id")
    assert device.num_qubits == sv1_profile.get("num_qubits")
    assert device._target_spec == sv1_profile.get("program_spec")
    assert device.simulator == sv1_profile.get("simulator")
    assert device.profile.get("experiment_type").value == "gate_model"


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_queue_depth(mock_aws_device, sv1_profile):
    """Test getting queue_depth BraketDevice"""
    mock_aws_device.queue_depth.return_value = QueueDepthInfo(
        quantum_tasks={QueueType.NORMAL: "19", QueueType.PRIORITY: "3"},
        jobs="0 (3 prioritized job(s) running)",
    )
    device = BraketDevice(sv1_profile)
    queue_depth = device.queue_depth()
    assert isinstance(queue_depth, int)


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_run_circuit_too_many_qubits(mock_aws_device, sv1_profile):
    """Test that run method raises exception when input circuit
    num qubits exceeds that of wrapped AWS device."""
    mock_aws_device.return_value = Mock()
    device = BraketDevice(sv1_profile)
    circuit = Circuit().h(0).cnot(0, device.num_qubits + 1)
    program = load_program(circuit)
    program.populate_idle_qubits()
    run_input = program.program
    with pytest.raises(ProgramValidationError):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            device.run(run_input)


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_batch_run(mock_aws_device, sv1_profile):
    """Test batch run method of BraketDevice."""
    mock_aws_device.return_value = MagicMock()
    mock_aws_device.return_value.__iter__.return_value = [MockTask("task1"), MockTask("task2")]
    mock_aws_device.return_value.status = "ONLINE"
    device = BraketDevice(sv1_profile)
    circuits = [Circuit().h(0).cnot(0, 1), Circuit().h(0).cnot(0, 1)]
    tasks = device.run(circuits, shots=10)
    assert isinstance(tasks, list)


@pytest.mark.parametrize(
    "available_time, expected",
    [
        (3600, "2024-01-01T01:00:00Z"),
        (1800, "2024-01-01T00:30:00Z"),
        (45, "2024-01-01T00:00:45Z"),
    ],
)
def test_availability_future_utc_datetime(available_time, expected):
    """Test calculating future utc datetime"""
    current_utc_datime = datetime.datetime(2024, 1, 1, 0, 0, 0)
    _, datetime_str = _calculate_future_time(available_time, current_utc_datime)
    assert datetime_str == expected


def test_device_availability_window(braket_provider, mock_sv1):
    """Test BraketDeviceWrapper avaliable output identical"""
    with (
        patch("qbraid.runtime.aws.provider.AwsDevice") as mock_aws_device,
        patch("qbraid.runtime.aws.device.AwsDevice") as mock_aws_device_2,
    ):
        mock_device = mock_sv1
        mock_device.is_available = False
        mock_aws_device.return_value = mock_device
        mock_aws_device_2.return_value = mock_device
        device = braket_provider.get_device(SV1_ARN)
        _, is_available_time, iso_str = device.availability_window()
        assert len(is_available_time.split(":")) == 3
        assert isinstance(iso_str, str)
        try:
            datetime.datetime.strptime(iso_str, "%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            pytest.fail("iso_str not in expected format")


@patch("qbraid.runtime.aws.BraketProvider")
def test_device_is_available(mock_provider):
    """Test device availability function."""
    provider = mock_provider.return_value

    mock_device_0 = Mock()
    mock_device_0.availability_window.return_value = (True, 0, 0)
    mock_device_0._device.is_available = True

    mock_device_1 = Mock()
    mock_device_1.availability_window.return_value = (False, 0, 0)
    mock_device_1._device.is_available = False

    provider.get_devices.return_value = [mock_device_0, mock_device_1]

    devices = provider.get_devices()
    for device in devices:
        is_available_bool, _, _ = device.availability_window()
        assert device._device.is_available == is_available_bool


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_transform_circuit_sv1(mock_aws_device, sv1_profile):
    """Test transform method for device with OpenQASM action type."""
    mock_aws_device.return_value = Mock()
    device = BraketDevice(sv1_profile)
    circuit = Circuit().h(0).cnot(0, 2)
    transformed_expected = Circuit().h(0).cnot(0, 1)
    transformed_circuit = device.transform(circuit)
    assert transformed_circuit == transformed_expected


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_transform_raises_for_mismatch(mock_aws_device, braket_circuit):
    """Test raising exception for mismatched action type AHS and program type circuit."""
    mock_aws_device.return_value = Mock()
    profile = TargetProfile(
        simulator=True,
        num_qubits=34,
        program_spec=ProgramSpec(Circuit),
        provider_name="Amazon Braket",
        device_id=SV1_ARN,
        experiment_type=ExperimentType.AHS,
    )
    device = BraketDevice(profile)
    with pytest.raises(DeviceProgramTypeMismatchError):
        device.transform(braket_circuit)


@pytest.mark.skipif(pytket_not_installed, reason="pytket not installed")
@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_ionq_transform(mock_aws_device):
    """Test converting Amazon Braket circuit to use only IonQ supprted gates"""
    mock_aws_device.return_value = Mock()
    profile = TargetProfile(
        simulator=False,
        num_qubits=11,
        program_spec=ProgramSpec(Circuit),
        provider_name="IonQ",
        device_id="arn:aws:braket:us-east-1::device/qpu/ionq/Harmony",
        experiment_type=ExperimentType.GATE_MODEL,
    )
    device = BraketDevice(profile)

    toffoli_circuit = Circuit().ccnot(0, 1, 2)
    transformed_circuit = device.transform(toffoli_circuit)
    assert isinstance(transformed_circuit, Circuit)
    assert toffoli_circuit.depth == 1
    assert transformed_circuit.depth == 11
    assert circuits_allclose(transformed_circuit, toffoli_circuit)


@patch("qbraid.runtime.aws.BraketProvider")
def test_device_submit_task_with_tags(mock_provider):
    """Test getting tagged quantum tasks."""
    provider = mock_provider.return_value
    provider.REGIONS = ["us-east-1", "us-west-1"]
    provider._get_default_region.return_value = "us-east-1"
    alt_regions = ["us-west-1"]

    mock_device = Mock()
    provider.get_device.return_value = mock_device

    mock_task_1 = Mock()
    mock_task_1.id = "task1"

    mock_task_2 = Mock()
    mock_task_2.id = "task2"

    circuit = Circuit().h(0).cnot(0, 1)
    mock_device.run.side_effect = [mock_task_1, mock_task_2]

    key, value1, value2 = "abc123", "val1", "val2"
    mock_task_1.tags = {key: value1}
    mock_task_2.tags = {key: value2}
    provider.get_tasks_by_tag.side_effect = [
        [mock_task_1.id, mock_task_2.id],
        [mock_task_1.id],
        [],
    ]

    task1 = mock_device.run(circuit, shots=2, tags={key: value1})
    task2 = mock_device.run(circuit, shots=2, tags={key: value2})

    assert set([task1.id, task2.id]) == set(provider.get_tasks_by_tag(key))
    assert len(provider.get_tasks_by_tag(key, values=[value1], region_names=["us-east-1"])) == 1
    assert len(provider.get_tasks_by_tag(key, region_names=alt_regions)) == 0


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_submit_with_partial_measurement_tags(mock_aws_device, sv1_profile):
    """Test that partial measurement qubits are properly tagged during submission."""
    # Mock the AWS device and its run_batch method
    mock_device_instance = MagicMock()
    mock_aws_device.return_value = mock_device_instance

    # Create a mock batch result
    mock_task = MagicMock()
    mock_task.id = "test_task_id"
    mock_device_instance.run.return_value = mock_task

    device = BraketDevice(sv1_profile)

    # Create a circuit with partial measurement qubits attribute
    circuit = Circuit().h(0).cnot(0, 1)
    circuit.partial_measurement_qubits = [0, 2]

    # Submit the circuit
    result = device.submit(circuit, shots=100)

    # Verify that run_batch was called with the correct tags
    expected_tags = {"partial_measurement_qubits": "0/2"}
    mock_device_instance.run.assert_called_once_with(circuit, tags=expected_tags, shots=100)

    # Verify the returned task
    assert isinstance(result, BraketQuantumTask)
    assert result.id == "test_task_id"


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_submit_without_partial_measurement_tags(mock_aws_device, sv1_profile):
    """Test submission without partial measurement qubits."""
    # Mock the AWS device and its run_batch method
    mock_device_instance = MagicMock()
    mock_aws_device.return_value = mock_device_instance

    # Create a mock batch result
    mock_task = MagicMock()
    mock_task.id = "test_task_id"
    mock_batch_result = MagicMock()
    mock_batch_result.tasks = [mock_task]
    mock_device_instance.run_batch.return_value = mock_batch_result

    device = BraketDevice(sv1_profile)

    circuit = Circuit().h(0).cnot(0, 1)
    result = device.submit(circuit, shots=100)

    # Verify that run_batch was called with empty tags
    mock_device_instance.run_batch.assert_called_once_with([circuit], shots=100)

    # Verify the returned task
    assert isinstance(result, BraketQuantumTask)
    assert result.id == "test_task_id"


@patch("qbraid.runtime.aws.device.AwsDevice")
def test_device_submit_batch_with_partial_measurements(mock_aws_device, sv1_profile):
    """Test batch submission with partial measurement tags."""
    # Mock the AWS device and its run_batch method
    mock_device_instance = MagicMock()
    mock_aws_device.return_value = mock_device_instance

    # Create mock batch results
    mock_task = MagicMock()
    mock_task.id = "test_task_id"
    mock_device_instance.run.return_value = mock_task

    device = BraketDevice(sv1_profile)

    # Create circuits with partial measurement qubits
    circuit1 = Circuit().h(0).cnot(0, 1)
    circuit1.partial_measurement_qubits = [0]
    circuit2 = Circuit().h(1).cnot(1, 2)
    circuit2.partial_measurement_qubits = [1, 2]
    results = device.submit([circuit1, circuit2], shots=100)

    # Verify the tags
    expected_tags1 = {"partial_measurement_qubits": "0"}
    expected_tags2 = {"partial_measurement_qubits": "1/2"}
    assert mock_device_instance.run.call_args_list[0][1]["tags"] == expected_tags1
    assert mock_device_instance.run.call_args_list[1][1]["tags"] == expected_tags2

    # Verify the returned tasks
    assert isinstance(results, list)
    assert len(results) == 2
    assert results[0].id == "test_task_id"
    assert results[1].id == "test_task_id"


@patch("qbraid.runtime.aws.job.BraketQuantumTask._get_partial_measurement_qubits_from_tags")
@patch("qbraid.runtime.aws.job.AwsQuantumTask")
def test_job_load_completed(mock_aws_quantum_task, mock_partial_measurements):
    """Test is terminal state method for BraketQuantumTask."""
    mock_partial_measurements.return_value = []
    circuit = Circuit().h(0).cnot(0, 1)
    mock_device = LocalSimulator()
    mock_job = mock_device.run(circuit, shots=10)
    mock_job.metadata = lambda: {
        "status": mock_job.state(),
        "quantumTaskArn": mock_job.id,
        "deviceArn": mock_device.name,
    }
    mock_aws_quantum_task.return_value = mock_job
    job = BraketQuantumTask(mock_job.id)
    assert job.metadata()["job_id"] == mock_job.id
    assert job.is_terminal_state()
    res = job.result()
    data: GateModelResultData = res.data
    assert data.measurements is not None


@pytest.mark.parametrize("position,expected", [(10, 10), (">2000", 2000)])
@patch("qbraid.runtime.aws.job.AwsQuantumTask")
def test_job_queue_position(mock_aws_quantum_task, position, expected):
    """Test getting queue_depth BraketDevice"""
    mock_queue_position_return = MagicMock()
    mock_queue_position_return.queue_position = position
    mock_aws_quantum_task.return_value.queue_position.return_value = mock_queue_position_return
    job = BraketQuantumTask("job_id")
    assert job.queue_position() == expected


@patch("qbraid.runtime.aws.job.AwsQuantumTask")
def test_job_queue_position_raises_version_error(mock_aws_quantum_task):
    """Test handling of AttributeError leads to raising AmazonBraketVersionError."""
    mock_aws_quantum_task.return_value.queue_position.side_effect = AttributeError
    job = BraketQuantumTask("job_id")
    with pytest.raises(AmazonBraketVersionError) as excinfo:
        job.queue_position()
    assert "Queue visibility is only available for amazon-braket-sdk>=1.56.0" in str(excinfo.value)


def test_job_raise_for_cancel_terminal():
    """Test raising an error when trying to cancel a completed/failed job."""
    with patch("qbraid.runtime.job.QuantumJob.is_terminal_state", return_value=True):
        job = BraketQuantumTask(SV1_ARN, task=Mock())
        with pytest.raises(JobStateError):
            job.cancel()


def test_result_measurements():
    """Test measurements method of BraketGateModelResultBuilder class."""
    mock_measurements = np.array([[0, 1, 1], [1, 0, 1]])
    mock_result = MagicMock()
    mock_result.measurements = mock_measurements
    result = BraketGateModelResultBuilder(mock_result)
    expected_output = np.array([[1, 1, 0], [1, 0, 1]])
    np.testing.assert_array_equal(result.measurements(), expected_output)


def test_result_get_counts():
    """Test get_counts method of BraketGateModelResultBuilder class."""
    mock_measurement_counts = {(0, 1, 1): 10, (1, 0, 1): 5}
    mock_result = MagicMock()
    mock_result.measurement_counts = mock_measurement_counts
    result = BraketGateModelResultBuilder(mock_result)
    expected_output = {"110": 10, "101": 5}
    assert result.get_counts() == expected_output


def test_result_measurements_with_partial_measurements():
    """Test measurements method with partial measurement qubits."""
    mock_measurements = np.array([[0, 1, 1, 0], [1, 0, 1, 1]])
    mock_result = MagicMock()
    mock_result.measurements = mock_measurements

    # Test with partial measurement qubits [0, 2] (first and third qubits)
    partial_measurement_qubits = [0, 2]
    result = BraketGateModelResultBuilder(mock_result, partial_measurement_qubits)

    # Should return only measurements for qubits 0 and 2, with reversed order
    expected_output = np.array([[1, 0], [1, 1]])  # Reversed: [[0, 1], [1, 1]]
    np.testing.assert_array_equal(result.measurements(), expected_output)


def test_result_get_counts_with_partial_measurements():
    """Test get_counts method with partial measurement qubits."""
    mock_measurement_counts = {"0110": 10, "1011": 5, "0111": 3}
    mock_result = MagicMock()
    mock_result.measurement_counts = mock_measurement_counts

    # Test with partial measurement qubits [0, 2] (first and third qubits)
    partial_measurement_qubits = [0, 2]
    result = BraketGateModelResultBuilder(mock_result, partial_measurement_qubits)

    # Should marginalize to qubits 0 and 2: "01": 10+3=13, "11": 5, then reverse
    expected_output = {"10": 13, "11": 5}
    assert result.get_counts() == expected_output


def test_result_builder_initialization_with_partial_measurements():
    """Test BraketGateModelResultBuilder initialization with partial measurements."""
    mock_result = MagicMock()
    partial_measurement_qubits = [0, 2, 3]

    result_builder = BraketGateModelResultBuilder(mock_result, partial_measurement_qubits)

    assert result_builder._result == mock_result
    assert result_builder.partial_measurement_qubits == partial_measurement_qubits


def test_result_builder_initialization_without_partial_measurements():
    """Test BraketGateModelResultBuilder initialization without partial measurements."""
    mock_result = MagicMock()

    result_builder = BraketGateModelResultBuilder(mock_result)

    assert result_builder._result == mock_result
    assert result_builder.partial_measurement_qubits is None


def test_get_default_region_error():
    """Test getting default region when an exception is raised."""
    with patch("qbraid.runtime.aws.provider.Session") as mock_boto_session:
        mock_boto_session.side_effect = Exception
        mock_boto_session.region_name.return_value = "not us-east-1"
        assert BraketProvider()._get_default_region() == "us-east-1"


def test_get_s3_default_bucket():
    """Test getting default S3 bucket."""
    with patch("qbraid.runtime.aws.provider.AwsSession") as mock_aws_session:
        mock_instance = mock_aws_session.return_value
        mock_instance.default_bucket.return_value = "default bucket"
        assert BraketProvider()._get_s3_default_bucket() == "default bucket"

        mock_instance.default_bucket.side_effect = Exception
        assert BraketProvider()._get_s3_default_bucket() == "amazon-braket-qbraid-jobs"


def test_get_quantum_task_cost():
    """Test getting quantum task cost."""
    task_mock = Mock()
    task_mock.arn = "fake_arn"
    task_mock.state.return_value = "COMPLETED"
    job = BraketQuantumTask("task_arn", task_mock)
    with patch("qbraid.runtime.aws.job.get_quantum_task_cost", return_value=0.1):
        assert job.get_cost() == 0.1


def test_built_runtime_profile_fail():
    """Test building a runtime profile with invalid device."""

    class FakeSession:
        """Fake Session for testing."""

        def __init__(self):
            self.region = "us-east-1"

        def get_device(self, arn):  # pylint: disable=unused-argument
            """Fake get_device method."""
            return {
                "deviceType": "SIMULATOR",
                "providerName": "Amazon Braket",
                "deviceCapabilities": "{}",
            }

    class FakeDevice:
        """Fake AWS Device for testing."""

        def __init__(self, arn, aws_session=None):
            self.arn = arn
            self.aws_session = aws_session or FakeSession()
            self.status = "ONLINE"
            self.properties = TestProperties()
            self.is_available = True

        @staticmethod
        def get_device_region(arn):  # pylint: disable=unused-argument
            """Fake get_device_region method."""
            return "us-east-1"

    provider = BraketProvider(
        aws_access_key_id="aws_access_key_id",
        aws_secret_access_key="aws_secret_access_key",
    )

    device = FakeDevice(arn=SV1_ARN, aws_session=FakeSession())
    with pytest.raises(QbraidError):
        assert provider._build_runtime_profile(device=device)


def test_braket_result_error():
    """Test Braket result decoding error."""
    task = MockTask("task1")
    braket_task = BraketQuantumTask("task1", task)
    with pytest.raises(ValueError):
        braket_task.result()


def test_braket_job_cancel():
    """Test Braket job cancel method."""
    task = MockTask("task2")
    braket_task = BraketQuantumTask("task2", task)
    assert braket_task.cancel() is None


@patch("boto3.client")
def test_get_partial_measurement_qubits_from_tags_with_partial_measurements(mock_boto_client):
    """Test retrieving partial measurement qubits from tags when they exist."""
    # Mock the Braket client response
    mock_client = MagicMock()
    mock_boto_client.return_value = mock_client
    mock_client.get_quantum_task.return_value = {"tags": {"partial_measurement_qubits": "0/2/3"}}

    task = MockTask("task1")
    braket_task = BraketQuantumTask("task1", task)

    # All measured qubits in the order they appear in results
    all_measurement_qubits = [0, 1, 2, 3, 4]

    result = braket_task._get_partial_measurement_qubits_from_tags(all_measurement_qubits)

    # Should return indices [0, 2, 3] in the all_measurement_qubits list
    expected_indices = [0, 2, 3]  # positions of qubits 0, 2, 3 in the list
    assert result == expected_indices

    # Verify the client was called correctly
    mock_client.get_quantum_task.assert_called_once_with(quantumTaskArn="task1")


@patch("boto3.client")
def test_get_partial_measurement_qubits_from_tags_without_partial_measurements(mock_boto_client):
    """Test retrieving partial measurement qubits from tags when they don't exist."""
    # Mock the Braket client response without partial measurement tags
    mock_client = MagicMock()
    mock_boto_client.return_value = mock_client
    mock_client.get_quantum_task.return_value = {"tags": {"other_tag": "some_value"}}

    task = MockTask("task1")
    braket_task = BraketQuantumTask("task1", task)

    all_measurement_qubits = [0, 1, 2]

    result = braket_task._get_partial_measurement_qubits_from_tags(all_measurement_qubits)

    # Should return None when no partial measurement tags exist
    assert result is None

    mock_client.get_quantum_task.assert_called_once_with(quantumTaskArn="task1")


@patch("boto3.client")
def test_get_partial_measurement_qubits_from_tags_complex_mapping(mock_boto_client):
    """Test partial measurement qubit mapping with non-contiguous qubits."""
    # Mock the Braket client response
    mock_client = MagicMock()
    mock_boto_client.return_value = mock_client
    mock_client.get_quantum_task.return_value = {"tags": {"partial_measurement_qubits": "1/3"}}

    task = MockTask("task1")
    braket_task = BraketQuantumTask("task1", task)

    # Measured qubits in a different order than circuit qubits
    all_measurement_qubits = [3, 1, 0, 2]  # qubits measured in this order

    result = braket_task._get_partial_measurement_qubits_from_tags(all_measurement_qubits)

    # Qubit 1 is at index 1, qubit 3 is at index 0 in all_measurement_qubits
    expected_indices = [1, 0]  # positions of qubits 1, 3 in the measurement results
    assert result == expected_indices


def test_get_tasks_by_tag_value_error():
    """Test getting tagged quantum tasks with invalid values."""
    with patch("qbraid.runtime.aws.provider.quantum_lib_proxy_state") as mock_proxy_state:
        mock_proxy_state.side_effect = ValueError

        provider = BraketProvider()

        try:
            result = provider.get_tasks_by_tag("key", ["value1", "value2"])
        except NoCredentialsError:
            pytest.skip("NoCredentialsError raised")

        assert isinstance(result, list)


def test_get_tasks_by_tag_qbraid_error():
    """Test getting tagged quantum tasks with jobs enabled."""
    with patch("qbraid.runtime.aws.provider.quantum_lib_proxy_state") as mock_proxy_state:
        mock_proxy_state.return_value = {"enabled": True}

        provider = BraketProvider()

        with pytest.raises(QbraidError):
            provider.get_tasks_by_tag("key", ["value1", "value2"])


@pytest.fixture
def mock_braket_provider_with_credentials():
    """Return a BraketProvider instance with mock credentials."""
    aws_provider = BraketProvider(
        aws_access_key_id="mock_access_key_id",
        aws_secret_access_key="mock_secret_access_key",
        aws_session_token="mock_session_token",
    )
    return aws_provider


@patch("builtins.hash", autospec=True)
def test_hash_method_creates_and_returns_hash(mock_hash, mock_braket_provider_with_credentials):
    """Test that the __hash__ method creates and returns a hash value."""
    mock_hash.return_value = 5555
    provider_instance = mock_braket_provider_with_credentials
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    mock_hash.assert_called_once_with(
        ("mock_access_key_id", "mock_secret_access_key", "mock_session_token")
    )
    assert result == 5555
    assert provider_instance._hash == 5555


def test_hash_method_returns_existing_hash(mock_braket_provider_with_credentials):
    """Test that the __hash__ method returns the existing hash value."""
    provider_instance = mock_braket_provider_with_credentials
    provider_instance._hash = 9876
    result = provider_instance.__hash__()  # pylint:disable=unnecessary-dunder-call
    assert result == 9876



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module containing models for schema-conformant ResultData classes.

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional, Type, TypeVar, Union, overload

import numpy as np

from qbraid.programs import ExperimentType

from .postprocess import counts_to_probabilities, normalize_data
from .schemas.experiment import (
    AhsExperimentMetadata,
    AnnealingExperimentMetadata,
    ExperimentMetadata,
    GateModelExperimentMetadata,
)

ResultDataType = TypeVar("ResultDataType", bound="ResultData")

KeyType = TypeVar("KeyType", str, int)

MeasCount = dict[KeyType, int]

MeasProb = dict[KeyType, float]


class ResultData(ABC):
    """Abstract base class for runtime results linked to a
    specific :class:`~qbraid.programs.ExperimentType`.
    """

    @property
    @abstractmethod
    def experiment_type(self) -> ExperimentType:
        """Returns the experiment type."""

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Converts the result data to a dictionary."""

    @classmethod
    @abstractmethod
    def from_dict(cls: Type[ResultDataType], data: dict[str, Any]) -> ResultDataType:
        """Creates a new ResultData instance from a dictionary."""

    @classmethod
    @overload
    def from_object(cls, model: GateModelExperimentMetadata, **kwargs) -> GateModelResultData: ...

    @classmethod
    @overload
    def from_object(cls, model: AnnealingExperimentMetadata, **kwargs) -> AnnealingResultData: ...

    @classmethod
    @overload
    def from_object(cls, model: AhsExperimentMetadata, **kwargs) -> AhsResultData: ...

    @classmethod
    def from_object(
        cls: Type[ResultDataType], model: ExperimentMetadata, **kwargs
    ) -> ResultDataType:
        """Creates a new ResultData instance from an ExperimentMetadata object."""
        return cls.from_dict(model.model_dump(**kwargs))


class GateModelResultData(ResultData):
    """Class for storing and accessing the results of a gate model quantum job."""

    def __init__(
        self,
        measurement_counts: Optional[Union[MeasCount, list[MeasCount]]] = None,
        measurements: Optional[Union[np.ndarray, list[np.ndarray]]] = None,
        measurement_probabilities: Optional[Union[MeasProb, list[MeasProb]]] = None,
        **kwargs,
    ):
        """Create a new GateModelResult instance."""
        self._measurement_counts = measurement_counts
        self._measurements = measurements
        self._measurement_probabilities = measurement_probabilities
        self._unscoped_data = kwargs
        self._cache = {
            "bin_nz": None,
            "bin_wz": None,
            "dec_nz": None,
            "dec_wz": None,
            "prob_bin_nz": None,
            "prob_bin_wz": None,
            "prob_dec_nz": None,
            "prob_dec_wz": None,
            "to_dict": None,
        }

    @property
    def experiment_type(self) -> ExperimentType:
        """Returns the experiment type."""
        return ExperimentType.GATE_MODEL

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> GateModelResultData:
        """Creates a new GateModelResult instance from a dictionary."""
        measurement_counts = data.pop("measurement_counts", data.pop("measurementCounts", None))
        measurements = data.pop("measurements", None)
        measurement_probabilities = data.pop("measurement_probabilities", None)

        if isinstance(measurements, list):
            measurements = np.array(measurements, dtype=object)

        return cls(
            measurement_counts=measurement_counts,
            measurements=measurements,
            measurement_probabilities=measurement_probabilities,
            **data,
        )

    @property
    def measurements(self) -> Optional[Union[np.ndarray, list[np.ndarray]]]:
        """Returns the measurements data of the run."""
        return self._measurements

    @property
    def measurement_counts(self) -> Optional[Union[MeasCount, list[MeasCount]]]:
        """Returns the histogram data of the run as passed in the constructor."""
        return self._measurement_counts

    def get_counts(
        self, include_zero_values: bool = False, decimal: bool = False
    ) -> Union[MeasCount, list[MeasCount]]:
        """
        Returns the histogram data of the run with optional zero values and binary/decimal keys.

        Args:
            include_zero_values (bool): Whether to include states with zero counts.
            decimal (bool): Whether to return counts with decimal keys (instead of binary).

        Returns:
            Union[dict[str, int], list[dict[str, int]]]: The histogram data.

        Raises:
            ValueError: If counts data is not available.
        """
        if self._measurement_counts is None:
            raise ValueError("Counts data is not available.")

        cache_key = f"{'dec' if decimal else 'bin'}_{'wz' if include_zero_values else 'nz'}"

        if self._cache[cache_key] is not None:
            return self._cache[cache_key]
        counts = normalize_data(
            self._measurement_counts, include_zero_values=include_zero_values, decimal=decimal
        )

        self._cache[cache_key] = counts

        return counts

    def get_probabilities(
        self, include_zero_values: bool = False, decimal: bool = False
    ) -> Union[MeasProb, list[MeasProb]]:
        """
        Returns the probabilities of the measurement outcomes based on counts.

        Args:
            include_zero_values (bool): Whether to include states with zero probabilities.
            decimal (bool): Whether to return probabilities with decimal keys (instead of binary).

        Returns:
            Union[MeasProb, list[MeasProb]: Probabilities of measurement outcomes.

        Raises:
            ValueError: If probabilities data is not available or
                        if measurement_probabilities is not a dictionary.
        """
        cache_key = f"prob_{'dec' if decimal else 'bin'}_{'wz' if include_zero_values else 'nz'}"

        if self._cache[cache_key] is not None:
            return self._cache[cache_key]

        if self._measurement_probabilities is not None:
            if not isinstance(self._measurement_probabilities, dict):
                raise ValueError("'measurement_probabilities' must be a dictionary.")

            probabilities = normalize_data(
                self._measurement_probabilities,
                include_zero_values=include_zero_values,
                decimal=decimal,
            )
        else:
            counts = self.get_counts(include_zero_values=include_zero_values, decimal=decimal)
            probabilities = counts_to_probabilities(counts)

        self._cache[cache_key] = probabilities

        return probabilities

    def to_dict(self) -> dict[str, Any]:
        """Converts the GateModelResulData instance to a dictionary."""
        if self._cache["to_dict"] is not None:
            return self._cache["to_dict"]

        counts = self.get_counts()
        probabilities = self.get_probabilities()
        shots = sum(counts.values())
        num_measured_qubits = len(next(iter(counts)))
        data = {
            "shots": shots,
            "num_measured_qubits": num_measured_qubits,
            "measurement_counts": counts,
            "measurement_probabilities": probabilities,
            "measurements": self._measurements,
        }
        self._cache["to_dict"] = data

        return data

    @staticmethod
    def _format_array(arr: np.ndarray) -> str:
        return f"array(shape={arr.shape}, dtype={arr.dtype})"

    def __repr__(self) -> str:
        if isinstance(self._measurements, np.ndarray):
            measurements_info = self._format_array(self._measurements)
        elif isinstance(self._measurements, list) and all(
            isinstance(arr, np.ndarray) for arr in self._measurements
        ):
            measurements_info = (
                "[" + ", ".join(self._format_array(arr) for arr in self._measurements) + "]"
            )
        else:
            measurements_info = self._measurements

        return (
            f"{self.__class__.__name__}("
            f"measurement_counts={self._measurement_counts}, "
            f"measurements={measurements_info}, "
            f"measurement_probabilities={self._measurement_probabilities}"
            f")"
        )


@dataclass
class AhsShotResult:
    """Class for storing the results of a single shot in an analog Hamiltonian simulation job."""

    success: bool
    pre_sequence: Optional[np.ndarray] = None
    post_sequence: Optional[np.ndarray] = None

    @staticmethod
    def _sequences_equal(seq1: Optional[np.ndarray], seq2: Optional[np.ndarray]) -> bool:
        """Helper function to compare two sequences, handling None values."""
        return (seq1 is None and seq2 is None) or (
            seq1 is not None and seq2 is not None and np.array_equal(seq1, seq2)
        )

    def __eq__(self, other):
        if not isinstance(other, AhsShotResult):
            return False
        return (
            self.success == other.success
            and self._sequences_equal(self.pre_sequence, other.pre_sequence)
            and self._sequences_equal(self.post_sequence, other.post_sequence)
        )

    def to_dict(self) -> dict[str, Union[bool, Optional[int]]]:
        """Convert the instance to a dictionary, converting numpy arrays to lists."""
        return {
            "success": self.success,
            "pre_sequence": self.pre_sequence.tolist() if self.pre_sequence is not None else None,
            "post_sequence": (
                self.post_sequence.tolist() if self.post_sequence is not None else None
            ),
        }

    @classmethod
    def from_dict(cls, data: dict[str, Union[bool, Optional[list[int]]]]) -> AhsShotResult:
        """Create an instance from a dictionary, converting lists to numpy arrays."""
        return cls(
            success=data["success"],
            pre_sequence=(
                np.array(data["pre_sequence"]) if data.get("pre_sequence") is not None else None
            ),
            post_sequence=(
                np.array(data["post_sequence"]) if data.get("post_sequence") is not None else None
            ),
        )


class AhsResultData(ResultData):
    """Class for storing and accessing the results of an analog Hamiltonian simulation job."""

    def __init__(
        self,
        measurement_counts: Optional[dict[str, int]] = None,
        measurements: Optional[list[AhsShotResult]] = None,
    ):
        self._measurement_counts = measurement_counts
        self._measurements = measurements

    @property
    def experiment_type(self) -> ExperimentType:
        """Returns the experiment type."""
        return ExperimentType.AHS

    @property
    def measurements(self) -> Optional[list[AhsShotResult]]:
        """Returns the measurements data of the run."""
        return self._measurements

    def get_counts(self) -> Optional[dict[str, int]]:
        """Returns the histogram data of the run."""
        return self._measurement_counts

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> AhsResultData:
        """Creates a new AhsResultData instance from a dictionary."""
        measurements = data.get("measurements")
        if measurements is not None:
            if not isinstance(measurements, list):
                raise ValueError("'measurements' must be a list or None.")
            if not all(isinstance(shot, dict) for shot in measurements):
                raise ValueError("Each item in 'measurements' must be a dictionary.")
            measurements = [AhsShotResult.from_dict(shot) for shot in measurements]

        return cls(
            measurements=measurements,
            measurement_counts=data.get("measurement_counts", data.get("measurementCounts")),
        )

    def to_dict(self) -> dict[str, Any]:
        """Converts the AhsResultData instance to a dictionary."""
        return {
            "measurement_counts": self._measurement_counts,
            "measurements": self._measurements,
        }

    def __eq__(self, other):
        if not isinstance(other, AhsResultData):
            return False

        if self._measurement_counts != other._measurement_counts:
            return False

        if self._measurements is None and other._measurements is None:
            return True
        if self._measurements is None or other._measurements is None:
            return False
        if len(self._measurements) != len(other._measurements):
            return False

        return all(s1 == s2 for s1, s2 in zip(self._measurements, other._measurements))

    def __repr__(self) -> str:
        """Return a string representation of the AhsResultData instance."""
        return f"{self.__class__.__name__}(measurement_counts={self._measurement_counts}, measurements={self._measurements})"  # pylint: disable=line-too-long


class AnnealingResultData(ResultData):
    """Class for storing and accessing the results of an annealing job."""

    def __init__(
        self, solutions: Optional[list[dict[str, Any]]] = None, num_solutions: Optional[int] = None
    ):
        self._solutions = solutions
        self._num_solutions = num_solutions

    @property
    def experiment_type(self) -> ExperimentType:
        """Returns the experiment type."""
        return ExperimentType.ANNEALING

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> AnnealingResultData:
        """Creates a new AnnealingResultData instance from a dictionary."""
        return cls(
            solutions=data.get("solutions"),
            num_solutions=data.get("num_solutions", data.get("numSolutions")),
        )

    @property
    def solutions(self) -> Optional[list[dict[str, Any]]]:
        """Returns the solutions data of the run."""
        return self._solutions

    @property
    def num_solutions(self) -> Optional[int]:
        """Returns the number of solutions."""
        if self._num_solutions is None and self._solutions is not None:
            self._num_solutions = len(self._solutions)
        return self._num_solutions

    def to_dict(self) -> dict[str, Any]:
        """Converts the AnnealingResultData instance to a dictionary."""
        return {
            "solutions": self.solutions,
            "num_solutions": self.num_solutions,
        }



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for providing transforamtions to ensure OpenQASM 3 compatibility
across various other quantum software frameworks.

"""
import math
import re
from functools import reduce

from openqasm3 import dumps, parse
from openqasm3.ast import Program, QuantumGate, Statement
from openqasm3.parser import QASM3ParsingError

from qbraid._logging import logger

GATE_DEFINITIONS = {
    "iswap": """
gate iswap _gate_q_0, _gate_q_1 {
  s _gate_q_0;
  s _gate_q_1;
  h _gate_q_0;
  cx _gate_q_0, _gate_q_1;
  cx _gate_q_1, _gate_q_0;
  h _gate_q_1;
}""",
    "sxdg": """
gate sxdg _gate_q_0 {
  s _gate_q_0;
  h _gate_q_0;
  s _gate_q_0;
}""",
    "cv": """
gate cv _gate_q_0, _gate_q_1 {
  ctrl @ sx _gate_q_0, _gate_q_1;
}""",
}


def _insert_gate_def(qasm3_str: str, defn: str) -> str:
    """Add single gate definition to an OpenQASM3 string."""

    lines = qasm3_str.splitlines()

    # Note: In future move gate definition after include statement.
    insert_index = 0
    for i, line in enumerate(lines):
        if "include" in line or "OPENQASM" in line:
            insert_index = i + 1
            break

    lines.insert(insert_index, defn.strip())

    return "\n".join(lines)


def insert_gate_def(qasm3_str: str, gate_name: str | list[str], force_insert: bool = False) -> str:
    """Add gate definitions to an OpenQASM3 string.

    Args:
        qasm3_str (str): QASM 3.0 string.
        gate_name (str | list[str]): Name(s) of the gate(s) to insert.
        force_insert (bool): If True, the gate definition will be added
            even if the gate is never referenced. Defaults to False.

    Returns:
        str: QASM 3.0 string with gate definition.

    Raises:
        ValueError: If the gate definition is not found.
    """
    # Normalize gate_name to a list
    gate_names = [gate_name] if isinstance(gate_name, str) else gate_name

    # Check for missing definitions
    missing = [name for name in gate_names if name not in GATE_DEFINITIONS]
    if missing:
        raise ValueError(
            f"Gate definitions not found for: {missing}. "
            f"Available gate definitions include: {set(GATE_DEFINITIONS.keys())}"
        )

    # Determine which gates to actually add
    gates_to_add = []
    if force_insert:
        gates_to_add = gate_names
    else:
        # Only add if gate name appears in QASM string
        for name in gate_names:
            if name in qasm3_str:
                gates_to_add.append(name)

        # If no gates referenced and force_insert is False, return original
        if not gates_to_add:
            return qasm3_str

    # Insert definitions
    for name in gates_to_add:
        defn = GATE_DEFINITIONS[name]
        qasm3_str = _insert_gate_def(qasm3_str, defn)

    return qasm3_str


def _normalize_case_insensitive_map(gate_mappings: dict[str, str]) -> dict[str, str]:
    """Normalize gate_mappings keys to lowercase and check for duplicates."""
    lowercased_map = {}
    for key, value in gate_mappings.items():
        lower_key = key.lower()
        if lower_key in lowercased_map:
            raise ValueError(
                f"Duplicate key detected after lowercasing: '{lower_key}'. Use case_sensitive=True."
            )
        lowercased_map[lower_key] = value
    return lowercased_map


def _replace_gate_in_statement(
    statement: Statement, gate_mappings: dict[str, str], case_sensitive: bool
) -> QuantumGate:
    """Replace gate names in a single statement if applicable."""
    if isinstance(statement, QuantumGate):
        gate_name = statement.name.name
        lookup_name = gate_name if case_sensitive else gate_name.lower()
        if lookup_name in gate_mappings:
            statement.name.name = gate_mappings[lookup_name]
    return statement


def _replace_gate_names(
    program: Program, gate_mappings: dict[str, str], case_sensitive: bool
) -> Program:
    """Replace occurrences of specified gate names in a openqasm3.ast.Program."""
    if not case_sensitive:
        gate_mappings = _normalize_case_insensitive_map(gate_mappings)

    statements = [
        _replace_gate_in_statement(stmnt, gate_mappings, case_sensitive)
        for stmnt in program.statements
    ]

    return Program(statements=statements, version=program.version)


def replace_gate_names(
    qasm: str, gate_mappings: dict[str, str], case_sensitive: bool = False
) -> str:
    """Replace occurrences of specified gate names in a QASM program string.

    Args:
        qasm (str): The QASM program as a string.
        gate_mappings (dict[str, str]): A dictionary mapping old gate names (keys)
            to new gate names (values).
        case_sensitive (bool): Whether the gate name replacement should be case-sensitive.
            Defaults to False.

    Returns:
        str: The modified QASM program with the gate names replaced.

    Raises:
        ValueError: If duplicate keys are found in the gate_mappings after lowercasing.
    """
    program = parse(qasm)

    program_out = _replace_gate_names(program, gate_mappings, case_sensitive)
    version_major = program_out.version.split(".")[0]
    qasm_out = dumps(program_out)

    if int(version_major) == 2:
        qasm_out = declarations_to_qasm2(qasm_out)

    return qasm_out


def add_stdgates_include(qasm_str: str) -> str:
    """Add 'include "stdgates.inc";' to the QASM string if it is missing."""
    if 'include "stdgates.inc";' in qasm_str:
        return qasm_str

    lines = qasm_str.splitlines()

    for i, line in enumerate(lines):
        if "OPENQASM" in line:
            lines.insert(i + 1, 'include "stdgates.inc";')
            break

    return "\n".join(lines)


def remove_stdgates_include(qasm: str) -> str:
    """Remove 'include "stdgates.inc";' from the QASM string."""
    return qasm.replace('include "stdgates.inc";', "")


def _evaluate_expression(match):
    """Helper function for simplifying arithmetic expressions within parentheses."""
    expr = match.group(1)
    try:
        simplified_value = eval(expr)  # pylint: disable=eval-used
        return f"({simplified_value})"
    except SyntaxError:
        return match.group(0)


def simplify_arithmetic_expressions(qasm_str: str) -> str:
    """Simplifies arithmetic expressions within parentheses in a QASM string."""

    pattern = r"\(([0-9+\-*/. ]+)\)"

    return re.sub(pattern, _evaluate_expression, qasm_str)


def convert_qasm_pi_to_decimal(qasm: str) -> str:
    """Convert all instances of 'pi' in the QASM string to their decimal value."""

    pattern = r"(?<![a-zA-Z])(\d*\.?\d*\s*[*/+-]\s*)?pi(\s*[*/+-]\s*\d*\.?\d*)?(?![a-zA-Z])"

    gate_defs = set()

    try:
        program = parse(qasm)

        for statement in program.statements:
            if isinstance(statement, QuantumGate):
                name = statement.name.name
                if "pi" in name:
                    gate_defs.add(name)
    except QASM3ParsingError as err:
        logger.debug("Failed to parse QASM program for pi conversion: %s", err)

    def replace_with_decimal(match: re.Match) -> str:
        expr: str = match.group()
        start = match.start()
        end = match.end()

        for gate_def in gate_defs:
            if gate_def in qasm[max(0, start - len(gate_def)) : end]:
                return expr  # pragma: no cover

        expr_with_pi_as_decimal = expr.replace("pi", str(math.pi))
        value = eval(expr_with_pi_as_decimal)  # pylint: disable=eval-used
        return str(value)

    return re.sub(pattern, replace_with_decimal, qasm)


def has_redundant_parentheses(qasm_str: str) -> bool:
    """Checks if a QASM string contains gate parameters with redundant parentheses."""

    pattern = r"\w+\(\(\s*[-+]?\d+(\.\d*)?\s*\)\)"
    if re.search(pattern, qasm_str):
        return True

    pattern_neg = r"\w+\(-\(\d*\.?\d+\)\)"
    if re.search(pattern_neg, qasm_str):
        return True

    return False


def simplify_parentheses_in_qasm(qasm_str: str) -> str:
    """Simplifies unnecessary parentheses around numbers in QASM strings."""

    lines = qasm_str.splitlines()
    simplified_lines = []

    pattern = r"\(\s*([-+]?\s*\d+(\.\d*)?)\s*\)"

    def simplify(match):
        return match.group(1).replace(" ", "")

    for line in lines:
        if has_redundant_parentheses(line):
            line = re.sub(pattern, simplify, line)
        simplified_lines.append(line)

    return "\n".join(simplified_lines)


def compose(*functions):
    """Compose multiple functions left to right."""

    def compose_two(f, g):
        return lambda x: g(f(x))

    return reduce(compose_two, functions, lambda x: x)


def normalize_qasm_gate_params(qasm: str) -> str:
    """Normalize the parameters of the gates in the QASM string using function composition."""
    transform_qasm = compose(
        convert_qasm_pi_to_decimal, simplify_arithmetic_expressions, simplify_parentheses_in_qasm
    )
    return transform_qasm(qasm)


def declarations_to_qasm2(qasm: str) -> str:
    """Converts QASM 3.0 qubit and bit declarations to QASM 2.0 qreg and creg declarations."""
    for declaration_type, replacement_type in [("qubit", "qreg"), ("bit", "creg")]:
        pattern = rf"{declaration_type}\[(\d+)\]\s+(\w+);"
        replacement = rf"{replacement_type} \2[\1];"
        qasm = re.sub(pattern, replacement, qasm)

    return qasm



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining QiskitBackend Class

"""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from qiskit.transpiler import PassManager
from qiskit_ibm_runtime import QiskitRuntimeService
from qiskit_ibm_runtime import SamplerV2 as Sampler

from qbraid.programs import load_program
from qbraid.runtime.device import QuantumDevice
from qbraid.runtime.enums import DeviceStatus
from qbraid.runtime.options import RuntimeOptions

from .job import QiskitJob

if TYPE_CHECKING:
    import qiskit
    import qiskit_ibm_runtime

    import qbraid.runtime.ibm


class QiskitBackend(QuantumDevice):
    """Wrapper class for IBM Qiskit ``Backend`` objects."""

    def __init__(
        self,
        profile: qbraid.runtime.TargetProfile,
        service: Optional[qiskit_ibm_runtime.QiskitRuntimeService] = None,
    ):
        """Create a QiskitBackend."""
        options = RuntimeOptions(pass_manager=None)
        options.set_validator("pass_manager", lambda x: x is None or isinstance(x, PassManager))

        super().__init__(profile=profile, options=options)
        self._service = service or QiskitRuntimeService()
        self._backend = self._service.backend(
            self.id, instance=getattr(self.profile, "instance", None)
        )

    def __str__(self):
        """String representation of the QiskitBackend object."""
        return f"{self.__class__.__name__}('{self._backend.name}')"

    def status(self):
        """Return the status of this Device.

        Returns:
            str: The status of this Device
        """
        if getattr(self.profile, "local", False):
            return DeviceStatus.ONLINE

        status = self._backend.status()
        if status.operational:
            if status.status_msg == "active":
                return DeviceStatus.ONLINE
            return DeviceStatus.UNAVAILABLE
        return DeviceStatus.OFFLINE

    def queue_depth(self) -> int:
        """Return the number of jobs in the queue for the ibm backend"""
        if getattr(self.profile, "local", False):
            return 0
        return self._backend.status().pending_jobs

    def transform(self, run_input: qiskit.QuantumCircuit) -> qiskit.QuantumCircuit:
        """Transpile a circuit for the device."""
        program = load_program(run_input)
        program.transform(self)
        return program.program

    # pylint: disable-next=arguments-differ
    def submit(
        self,
        run_input: qiskit.QuantumCircuit | list[qiskit.QuantumCircuit],
        shots: int | None = None,
    ) -> qbraid.runtime.ibm.QiskitJob:
        """Runs circuit(s) on qiskit backend via :meth:`~SamplerV2.run`.

        Uses the :meth:`SamplerV2.execute` method to create a
        :class:`~qbraid.runtime.ibm.QiskitJob`, and return the result.

        Args:
            run_input: A circuit object to run on the IBM device.
            shots (int, optional): The number of times to run the task on the device. If None,
                number of shots is determined by the sampler.

        Keyword Args:
            shots (int, optional): The number of times to run the task on the device. If None,
                number of shots is determined by the sampler.

        Returns:
            qbraid.runtime.ibm.QiskitJob: The job like object for the run.

        """
        backend = self._backend
        sampler = Sampler(mode=backend)
        pubs = run_input if isinstance(run_input, list) else [run_input]
        job = sampler.run(pubs, shots=shots)
        return QiskitJob(job.job_id(), job=job, device=self)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module defining Amazon Braket conversion extras.

"""
from __future__ import annotations

from typing import TYPE_CHECKING

from qbraid_core._import import LazyLoader

from qbraid.transpiler.annotations import requires_extras

qiskit_braket_provider = LazyLoader("qiskit_braket_provider", globals(), "qiskit_braket_provider")
pytket_braket = LazyLoader("pytket_braket", globals(), "pytket.extensions.braket")

if TYPE_CHECKING:
    import braket.circuits
    import pytket.circuit
    import qiskit.circuit


@requires_extras("qiskit_braket_provider")
def braket_to_qiskit(circuit: braket.circuits.Circuit) -> qiskit.circuit.QuantumCircuit:
    """Return a Qiskit quantum circuit from a Braket quantum circuit.

    Args:
        circuit (Circuit): Braket quantum circuit

    Returns:
        QuantumCircuit: Qiskit quantum circuit
    """
    return qiskit_braket_provider.providers.adapter.to_qiskit(circuit)


@requires_extras("pytket.extensions.braket")
def braket_to_pytket(circuit: braket.circuits.Circuit) -> pytket.circuit.Circuit:
    """Returns a pytket circuit equivalent to the input Amazon Braket circuit.

    Args:
        circuit (braket.circuits.Circuit): Braket circuit to convert to a pytket circuit.

    Returns:
        pytket.circuit.Circuit: PyTKET circuit object equivalent to input Braket circuit.
    """
    return pytket_braket.braket_convert.braket_to_tk(circuit)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Module for plotting qBraid transpiler quantum program conversion graphs.

"""
from __future__ import annotations

import math
import random
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, Optional

import rustworkx as rx
from qbraid_core._import import LazyLoader
from rustworkx.visualization import mpl_draw

from qbraid.programs.experiment import ExperimentType
from qbraid.programs.registry import is_registered_alias_native

if TYPE_CHECKING:
    import matplotlib.pyplot

    import qbraid.runtime
    import qbraid.transpiler

plt: matplotlib.pyplot = LazyLoader("plt", globals(), "matplotlib.pyplot")

transpiler: qbraid.transpiler = LazyLoader("transpiler", globals(), "qbraid.transpiler")


def _validate_colors(colors: Optional[dict[str, str]] = None) -> dict[str, str]:
    """Validate and return the colors dictionary."""
    default_colors = {
        "target_node_outline": "red",
        "qbraid_node": "lightblue",
        "external_node": "lightgray",
        "qbraid_edge": "gray",
        "external_edge": "blue",
        "extras_edge": "darkgrey",
    }

    if colors is None:
        return default_colors

    extra_keys = set(colors.keys()) - set(default_colors.keys())
    if extra_keys:
        raise ValueError(f"Unexpected keys in 'colors': {extra_keys}")

    for key, value in default_colors.items():
        if key not in colors:
            colors[key] = value

    return colors


def plot_conversion_graph(  # pylint: disable=too-many-arguments
    graph: qbraid.transpiler.ConversionGraph,
    title: Optional[str] = "qBraid Quantum Program Conversion Graph",
    legend: bool = False,
    seed: Optional[int] = None,
    node_size: int = 1200,
    min_target_margin: int = 18,
    ax_margins: float = 0.1,
    show: bool = True,
    save_path: Optional[str | Path] = None,
    colors: Optional[dict[str, str]] = None,
    edge_labels: bool = False,
    experiment_type: Optional[ExperimentType | Iterable[ExperimentType]] = None,
    target_nodes: Optional[Iterable[str]] = None,
    **kwargs,
) -> None:
    """
    Plot the conversion graph using matplotlib. The graph is displayed using node
    and edge color conventions, with options for a title, legend, and figure saving.

    Args:
        graph (ConversionGraph): The directed conversion graph to be plotted.
        title (str | None): Title of the plot. Defaults to
            'qBraid Quantum Program Conversion Graph'.
        legend (bool): If True, display a legend on the graph. Defaults to False.
        seed (int | None): Seed for the node layout algorithm. Useful for consistent
            positioning. Defaults to None.
        node_size (int): Size of the nodes. Defaults to 1200.
        min_target_margin (int): Minimum target margin for edges. Defaults to 18.
        ax_margins (float): Padding added (as a fraction of data range) to auto-scaled
            axis limits. Defaults to 0.1.
        show (bool): If True, display the figure. Defaults to True.
        save_path (str | Path | None): Path to save the figure. If None, figure is not saved.
        colors (dict[str, str] | None): Node and edge colors with keys 'target_node_outline',
            'qbraid_node', 'external_node', 'qbraid_edge', 'external_edge'. Defaults to None.
        edge_labels (bool): If True, display edge weights as labels. Defaults to False.
        experiment_type (ExperimentType | Iterable[ExperimentType] | None): Filter the
            graph by experiment type. Defaults to None, meaning all experiment types are included.
        target_nodes (Iterable[str] | None): Nodes to be outlined in the plot. Defaults to None.

    Returns:
        None
    """
    colors = _validate_colors(colors)

    if experiment_type:
        graph = graph.subgraph(experiment_type)

    ncolors = [
        colors["qbraid_node"] if is_registered_alias_native(node) else colors["external_node"]
        for node in graph.nodes()
    ]

    conversion_dict = {
        (conversion.source, conversion.target): conversion for conversion in graph.conversions()
    }
    conversions_ordered = [
        conversion_dict[(graph.get_node_data(edge[0]), graph.get_node_data(edge[1]))]
        for edge in graph.edge_list()
        if (graph.get_node_data(edge[0]), graph.get_node_data(edge[1])) in conversion_dict
    ]
    ecolors = [
        (
            colors["qbraid_edge"]
            if graph.get_edge_data(
                graph._node_alias_id_map[edge.source], graph._node_alias_id_map[edge.target]
            )["native"]
            else colors["extras_edge"] if len(edge._extras) > 0 else colors["external_edge"]
        )
        for edge in conversions_ordered
    ]

    rustworkx_version = rx.__version__  # pylint: disable=no-member
    if len(set(ecolors)) > 1 and rustworkx_version in ["0.15.0", "0.15.1"]:
        warnings.warn(
            "Detected multiple edge colors, which may not display correctly "
            "due to a known bug in rustworkx>=0.15.0,<0.16.0 "
            "(see: https://github.com/Qiskit/rustworkx/issues/1308). "
            "To avoid this issue, please upgrade to rustworkx>=0.16.0.",
            UserWarning,
        )
    seed = seed or random.randint(1, 999)
    k = kwargs.pop("k", max(1 / math.sqrt(len(graph.nodes())), 3))
    pos = rx.spring_layout(graph, seed=seed, k=k, **kwargs)
    kwargs = {}
    if edge_labels:
        kwargs["edge_labels"] = lambda edge: round(edge["weight"], 2)

    if target_nodes:
        edgecolors = [
            colors["target_node_outline"] if node in target_nodes else color
            for node, color in zip(graph.nodes(), ncolors)
        ]

        kwargs["edgecolors"] = edgecolors

    plt.ioff()  # Disable interactive mode

    fig = plt.figure()
    ax = fig.add_subplot(111)

    mpl_draw(
        graph,
        pos,
        ax=ax,
        node_color=ncolors,
        edge_color=ecolors,
        node_size=node_size,
        with_labels=True,
        labels=str,
        min_target_margin=min_target_margin,
        **kwargs,
    )

    ax.margins(ax_margins, ax_margins)

    if title:
        plt.title(title)
    plt.axis("on")

    if legend:
        legend_info = [
            (
                ("Target - Node", "o", colors["target_node_outline"], "white", None)
                if target_nodes
                else None
            ),
            (
                ("qBraid - Node", "o", None, colors["qbraid_node"], None)
                if colors["qbraid_node"] in ncolors
                else None
            ),
            (
                ("External - Node", "o", None, colors["external_node"], None)
                if colors["external_node"] in ncolors
                else None
            ),
            (
                ("qBraid - Edge", None, None, colors["qbraid_edge"], "-")
                if colors["qbraid_edge"] in ecolors
                else None
            ),
            (
                ("Extras - Edge", None, None, colors["extras_edge"], "-")
                if colors["extras_edge"] in ecolors
                else None
            ),
            (
                ("External - Edge", None, None, colors["external_edge"], "-")
                if colors["external_edge"] in ecolors
                else None
            ),
        ]

        legend_info = [entry for entry in legend_info if entry is not None]

        legend_elements = [
            plt.Line2D(
                [0],
                [0],
                marker=marker,
                color="w" if marker else color,
                label=label,
                markersize=10 if marker else None,
                markeredgecolor=edgecolor if marker else None,
                markerfacecolor=color if marker else None,
                linestyle=linestyle,
                linewidth=2 if linestyle else None,
            )
            for label, marker, edgecolor, color, linestyle in legend_info
        ]
        plt.legend(handles=legend_elements, loc="best")

        text = f"seed: {seed}"

        plt.text(
            x=0.01,
            y=0.01,
            s=text,
            transform=plt.gca().transAxes,
            fontsize=8,
            color="gray",
            ha="left",
            va="bottom",
        )

    fig.tight_layout()

    if save_path:
        plt.savefig(save_path, bbox_inches="tight", dpi=300)

    if show:
        plt.show(block=True)  # Explicit blocking show


def plot_runtime_conversion_scheme(device: qbraid.runtime.QuantumDevice, **kwargs) -> None:
    """
    Plot the runtime conversion scheme for a given quantum device.

    Args:
        device (QuantumDevice): The quantum device for which to plot the conversion scheme.
        **kwargs: Additional keyword arguments to pass to the plotting function.

    Returns:
        None
    """
    if device.profile.program_spec:
        device.scheme.reset_graph()
        device.scheme.update_graph_for_target(device.profile.program_spec)

    graph = device.scheme.conversion_graph
    experiment_type = device.profile.experiment_type
    target_spec = device.profile.program_spec

    target_nodes = (
        {spec.alias for spec in (target_spec if isinstance(target_spec, list) else [target_spec])}
        if target_spec
        else None
    )

    title = kwargs.pop("title", f"Runtime Conversion Scheme for {str(device)}")

    plot_conversion_graph(
        graph,
        title=title,
        experiment_type=experiment_type,
        target_nodes=target_nodes,
        **kwargs,
    )



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Unit test for the graph-based transpiler

"""
import unittest.mock

import braket.circuits
import pytest

from qbraid.programs import register_program_type
from qbraid.transpiler.converter import _warn_if_unsupported, transpile
from qbraid.transpiler.edge import Conversion
from qbraid.transpiler.exceptions import ConversionPathNotFoundError, NodeNotFoundError
from qbraid.transpiler.graph import ConversionGraph


def test_unsupported_target_package():
    """Test that an error is raised if target package is not supported."""
    with pytest.raises(NodeNotFoundError):
        transpile(braket.circuits.Circuit(), "alice")


def test_transpile_bad_source():
    """Test raising ValueError for bad source package"""

    def mock_has_node(node):
        """Mock function for has_node"""
        return not node == "fake"

    class FakeClass:
        """Fake class for testing"""

    register_program_type(FakeClass, "fake")
    with unittest.mock.patch(
        "qbraid.transpiler.ConversionGraph.has_node", side_effect=mock_has_node
    ):
        with pytest.raises(NodeNotFoundError):
            transpile(FakeClass(), "cirq")


def test_warning():
    """Test that a warning is raised if the program type is not supported."""
    with pytest.warns(UserWarning):
        _warn_if_unsupported("test", "bad")


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_raise_no_conversion_path_found(bell_circuit):
    """Test raising exception when no conversion path is found"""
    qiskit_circuit, _ = bell_circuit
    conversions = [
        Conversion("cirq", "braket", lambda x: x),
        Conversion("cirq", "qiskit", lambda x: x),
    ]
    graph = ConversionGraph(conversions)
    with pytest.raises(ConversionPathNotFoundError):
        transpile(qiskit_circuit, "braket", conversion_graph=graph)


@pytest.mark.parametrize("bell_circuit", ["qiskit"], indirect=True)
def test_raise_no_conversion_path_found_max_depth(bell_circuit):
    """Test raising exception when no conversion path is found when a conversion path
    exists but does not meet the max_depth requirement."""
    qiskit_circuit, _ = bell_circuit
    with pytest.raises(ConversionPathNotFoundError):
        transpile(qiskit_circuit, "braket", max_path_depth=1, require_native=True)



# Copyright (C) 2024 qBraid
#
# This file is part of the qBraid-SDK
#
# The qBraid-SDK is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for the qBraid-SDK, as per Section 15 of the GPL v3.

"""
Pennylane conversions

.. currentmodule:: qbraid.transpiler.conversions.pennylane

Functions
----------

.. autosummary::
   :toctree: ../stubs/

   pennylane_to_qasm2

"""
from .pennylane_to_qasm2 import pennylane_to_qasm2

__all__ = ["pennylane_to_qasm2"]
